import copy
import numpy as np
import warnings
import matplotlib.pyplot as plt
import yaml
import math
from scipy.spatial.distance import pdist
from scipy.spatial.transform import Rotation as R
from ..utils.transform.points_transforms import (
    rotate_pts_by_vector,
    transform_displace_set,
    planar_rotation
)
from ..utils.data_format.visualisation import format_coordinates, set_colorplot
from ..utils.visualisation.matplotlib_plots import (
    add_ax_scatter,
    draw1nomral_segment,
    draw_nomral_segments,
)
from ..utils.transform.cif_builder import create_instance_label
from ..utils.transform.structural_integrity import xmersubset_byclustering
from ..utils.data_format.structural_format import builder_format
from ..utils.sample.arrays import sample_array

class LabeledInstance:
    def __init__(self):
        """
        Initialize a LabeledInstance object with default parameters and attributes.
        """
        self.params = {}
        self.params["ref_point"] = None
        self.params["scale"] = 1e-10
        self.source = {}
        self.source["targets"] = None
        # self.source["reference_pt"] = None
        # self.source["scale"] = None
        # self.source["axis"] = None
        self.source["info"] = None
        self.labels = {}
        self.labelnames = list()
        # instance related
        self.emitters = []
        self.axis = dict()
        self.plotting_params = dict()
        self.radial_hindance = None
        self.structural_integrity = False
        self.structural_integrity_params = dict()
        # Contains the subset of epitopes after structural_integrity calculation
        self.structural_integrity_target_normals = None
        self.fluo2labels = []
        self.status = dict(source=False, labels=False)
        # attributes for sequencial labelling
        self.sequential_labelling = False
        self.primary = {}
        self.primary["targets"] = {}
        self.primary["reference_pt"] = None
        self.primary["scale"] = None
        self.primary["axis"] = None
        self.primary["info"] = None
        #
        self.secondary = {}

    def set_params(self, **kwargs):
        """
        Set parameters for the LabeledInstance using keyword arguments.

        Parameters
        ----------
        **kwargs
            Parameter names and values to set.
        """
        for key, value in kwargs.items():
            self.params[key] = value

    def _set_radial_hindrance(self, hindrance):
        self.radial_hindance = hindrance

    # methods to define source entity
    def _set_source_targets(self, targets: dict):
        # keys are the label agent used to define such targets
        self.source["targets"] = targets
        # estimate particle size
        hindrance = 0
        for tgt in self.source["targets"].keys():
            epitopes = self._get_source_coords_normals(tgt)["coordinates"]
            ref_and_epitopes = np.vstack(
                [self.get_ref_point().reshape(1, 3), epitopes]
            )
            max_dist = np.max(pdist(ref_and_epitopes))
            if max_dist > hindrance:
                hindrance = max_dist
        self.radial_hindance = hindrance

    def _set_source_reference(self, ref):
        self.source["reference_pt"] = ref

    def _set_source_scale(self, scale):
        self.source["scale"] = scale

    def _set_source_axis(self, axis: dict):
        self.source["axis"] = axis

    def _set_source_info(self, info: str):
        self.source["info"] = info

    def load_source(
        self,
        targets: dict,
        reference_point: np.ndarray,
        scale: float,
        axis=dict(pivot=np.array([0, 0, 0]), direction=np.array([0, 0, 1])),
        info="NA",
        **kwargs,
    ):
        """
        Set the ideal label locations and information for label targeting.

        Parameters
        ----------
        targets : dict
            Dictionary with target locations per label entity.
        reference_point : numpy.ndarray
            Reference point shared by all targets.
        scale : float
            Scale of the coordinates.
        axis : dict, optional
            Axis definition with 'pivot' and 'direction'.
        info : str, optional
            Additional information string.
        **kwargs
            Additional keyword arguments.
        """
        if scale != self.params["scale"]:
            scaling_factor = scale / self.params["scale"]
            for tgt in targets.keys():
                if targets[tgt]["normals"] is not None:
                    targets[tgt]["normals"] *= scaling_factor
                targets[tgt]["coordinates"] *= scaling_factor
            reference_point *= scaling_factor
            axis["pivot"] *= scaling_factor
            axis["direction"] *= scaling_factor
        self._set_ref_point(reference_point)
        self._set_source_targets(dict(targets))  # making an explicit copy
        # self._set_source_reference(reference_point)
        self._set_source_scale(scale)
        # self._set_source_axis(axis)
        self.axis = copy.copy(axis)
        self._set_source_info(info)
        self.status["source"] = True
        self.axis_reset = copy.copy(axis)

    def _get_source_target_normals(self, target_name):
        if self.source["targets"][target_name]["normals"] is None:
            return None
        else:
            return self.source["targets"][target_name]["normals"]

    def _get_source_coords_normals(self, target_name):
        # output is a dictionary with keys "coordinates" and "normals"
        return self.source["targets"][target_name]

    def _get_source_target_names(self):
        return self.source["targets"].keys()

    def _get_source_target_label(self, target_name):
        if target_name in self.labels.keys():
            return self.labels[target_name]
        else:
            return None

    def _get_source_parameter(self, key):
        return self.source[key]

    def load_label_file(self, label_yaml):
        """
        Load label definitions from a YAML file and add them to the instance.

        Parameters
        ----------
        label_yaml : str
            Path to the YAML file containing label definitions.
        """
        with open(label_yaml, "r") as f:
            label_params = yaml.safe_load(f)
        targets = dict()
        for lab in label_params["labels"]:
            # print(lab)
            fluorophore = lab["fluorophore"]
            coordinates = None
            if "coordinates" in lab.keys():
                # print(lab["coordinates"])
                coordinates = np.array(lab["coordinates"])
                # print(coordinates, type(coordinates))
            targets[fluorophore] = coordinates
            self.load_label(
                targets=targets,
                label_name=lab["label_name"],
                labelling_efficiency=lab["labeling_efficiency"],
            )

    # methods to specify labels
    def load_label(
        self,
        targets: dict,
        scale: float = 1e-10,
        axis=None,
        label_name="NA",
        labelling_efficiency: float = 1.0,
        minimal_distance: float = 0.0,
        secondary=False,
        info="NA",
        **kwargs,
    ):
        """
        Add a label to the instance.

        Parameters
        ----------
        targets : dict
            Dictionary with fluorophore names as keys and emitter coordinates as values.
        scale : float, optional
            Scale of the label. Default is 1e-9.
        axis : dict, optional
            Axis definition for the label.
        label_name : str, optional
            Name of the label.
        labelling_efficiency : float, optional
            Labelling efficiency. Default is 1.0.
        minimal_distance : float, optional
            Minimal distance between emitters. Default is 0.0.
        secondary : bool, optional
            If True, label is secondary. Default is False.
        info : str, optional
            Additional information.
        **kwargs
            Additional keyword arguments.
        """
        # Set default axis if None is provided or if axis has None values
        if axis is None:
            axis = dict(pivot=[0, 0, 0], direction=[0, 0, 1])
        else:
            # Ensure axis has valid values, not None
            if axis.get("pivot") is None:
                axis["pivot"] = [0, 0, 0]
            if axis.get("direction") is None:
                axis["direction"] = [0, 0, 1]

        # should check if the name is different
        label_name = label_name
        self.labelnames.append(label_name)
        for (
            keys,
            values,
        ) in targets.items():  # assumption: theres only one key and value
            fluorophore = keys
            emitters = values
        if emitters is None:
            label_type = "direct"
        else:
            label_type = "indirect"
            # add minimal distance to radial hindrance
            probe_max_dist = np.max(pdist(emitters))
            new_hindrance = self.radial_hindance + 2 * probe_max_dist
            if new_hindrance > self.radial_hindance:
                self.radial_hindance = new_hindrance
        colour = set_colorplot(self.plotting_params)
        # print(f"colour assigned to label: {colour}")
        plt_params = dict(
            plotsize=20, plotalpha=1, plotmarker="o", plotcolour=colour
        )
        self.plotting_params[label_name] = plt_params
        label_params = dict(
            emitters=emitters,
            scale=scale,
            axis=axis,
            labelling_efficiency=labelling_efficiency,
            minimal_distance=minimal_distance,
            label_type=label_type,
            fluorophore=fluorophore,
        )
        for key, val in kwargs.items():
            label_params[key] = val
        if secondary:
            self.secondary[label_name] = label_params
        else:
            self.labels[label_name] = label_params
            self.status["labels"] = True

    def _get_labels(self):

        return self.labels

    def get_label_names(self):
        """
        Get the list of label names.

        Returns
        -------
        list of str
            Names of all labels in the instance.
        """
        return self.labelnames

    def _get_label_fluorophore(self, label_name):
        return self.labels[label_name]["fluorophore"]

    def _get_label_emitters(self, label_name):
        # by default we consider each label has
        # a single type of fluorophore conjugated to it
        return self.labels[label_name]["emitters"]

    def _get_label_efficiency(self, label_name):
        return self.labels[label_name]["labelling_efficiency"]

    def _get_label_axis(self, label_name):
        return self.labels[label_name]["axis"]

    def _get_label_minimal_dist(self, label_name):
        return self.labels[label_name]["minimal_distance"]

    def _get_label_type(self, label_name):
        return self.labels[label_name]["label_type"]

    def _get_label_plotting_params(self, label_name):
        if label_name in self.plotting_params.keys():
            return self.plotting_params[label_name]
        else:
            default = {
                "plotsize": 1,
                "plotalpha": 1,
                "plotmarker": "o",
                "plotcolour": "k",
            }
            return default

    # methods to change label parameters
    def _change_label_fluorophore(self, label_name, newfluorophore):
        pass

    def _change_label_efficiency(self, label_name, newefficiency):
        self.labels[label_name]["labelling_efficiency"] = newefficiency

    # methods to generate a labelled instance from a target
    def _label_source_target(self, target_name):
        """
        Given a label name, it will consider the type of labeling
        and will create an instance of emitter locations following the source
        and the label characteristics

        Method inteded to be called for each of the targets in the Source
        """
        # The target name should correspond to the label name
        # print(f"Generating instance for label {target_name}")

        if (
            self._get_source_target_normals(target_name) is None
        ):  # if no normals exist this is true
            target_type = "direct"
            # print(f"Label has no normals defined. Using as direct labelling")
        else:
            target_type = "indirect"
            # print(f"Label has normals defined. Using as indirect labelling")
        target_normals = self._get_source_coords_normals(target_name)
        ###### print(f"target_normals before breaking: {target_normals}")
        label4target = self._get_source_target_label(target_name)
        # at this point we can sample pairs of target_normals to model structural_integrity
        if self.structural_integrity:
            structural_integrity_target_normals = self._model_structural_integrity(target_normals)
            target_normals = copy.copy(structural_integrity_target_normals)
        else:
            structural_integrity_target_normals = None
        ##### print(f"target_normals after breaking: {target_normals}")
        # print(target_normals, label4target)
        labelling_realisation = np.array([])
        plotting_params = self._get_label_plotting_params(target_name)
        fluorophore_name = self._get_label_fluorophore(target_name)
        labelling_realisation_vectors = None
        if target_normals["coordinates"].shape[0] > 0:
            if label4target is not None:
                labelling_realisation, n_epitopes, labelling_realisation_vectors = (
                    create_instance_label(
                        target_normals, target_type, label4target
                    )
                )
                plotting_params = self._get_label_plotting_params(target_name)
                fluorophore_name = self._get_label_fluorophore(target_name)
            else:
                warnings.warn(
                    f'No Label correspond to target "{target_name}". '
                    "Target will no be considered"
                )
            # generate the emitters positions
            # create_instance_label is capable of cosidering direct and indirect labelling
            # Once created, define the instance
        else:
            n_epitopes = 0
        return (
            labelling_realisation,
            labelling_realisation_vectors,
            fluorophore_name,
            plotting_params,
            structural_integrity_target_normals,
            n_epitopes
        )

    def _generate_instance_constructor(self):
        targets_labeled_instance = {}
        targets_labeled_instance_vectors = {}
        label_fluorophore = {}
        plotting_params = {}
        # for each target create the emitters
        for target_name in self._get_source_target_names():
            # print(target_name)
            self.structural_integrity_target_normals = None
            (
                emitters,
                labelling_realisation_vectors,
                fluorophore_name,
                plotting_par,
                structural_integrity_target_normals,
                n_epitopes
            ) = self._label_source_target(target_name)
            if structural_integrity_target_normals is not None:
                self.structural_integrity_target_normals = structural_integrity_target_normals
            # print(f"Emitters for target {target_name}")
            # print(emitters)
            targets_labeled_instance[target_name] = emitters
            targets_labeled_instance_vectors[target_name] = (
                labelling_realisation_vectors
            )
            label_fluorophore[target_name] = fluorophore_name
            plotting_params[target_name] = plotting_par
        # target_labeled_instance
        # then crete the instance constructor dictionary
        instance_constructor = dict(
            emitters_dictionary=targets_labeled_instance,
            emitters_vectors=targets_labeled_instance_vectors,
            ref_point=self.get_ref_point(),
            scale=self.get_scale(),
            axis=self.axis,
            label_fluorophore=label_fluorophore,
            plotting_params=plotting_params,
            n_epitopes=n_epitopes
        )
        return instance_constructor

    def _generate_primary_epitopes(
        self, emitters_dictionary, emitters_vectors, **kwargs
    ):
        # label the soruces using only the primary
        # create another sequential source
        for target_name in emitters_dictionary.keys():
            col_vec = np.array([i for i in emitters_vectors[target_name]])
            self.primary["targets"][target_name] = dict(
                coordinates=emitters_dictionary[target_name], normals=col_vec
            )

    def _label_primary(self, target_name):
        emitters = None
        target_normals = self.primary["targets"][target_name]
        if target_name in self.secondary.keys():
            label4target = self.secondary[target_name]
            # print(target_normals.keys())
            print(label4target)
            emitters, n_epitopes, labelling_realisation_vectors = create_instance_label(
                target_normals, "indirect", label4target
            )
            return emitters
        else:
            print(f"No secondary for {target_name}")

    def _generate_secondary_labelling(self):
        targets_labeled_instance = {}
        label_fluorophore = {}
        plotting_params = {}
        # for each target create the emitters
        for target_name in self.primary["targets"].keys():
            emitters = self._label_primary(target_name)
            targets_labeled_instance[target_name] = emitters
            # fluorophore_name = self._get_label_fluorophore(target_name)
            label_fluorophore[target_name] = self.secondary[target_name][
                "fluorophore"
            ]
            plotting_params[target_name] = self.plotting_params[target_name]

        instance_constructor = dict(
            emitters_dictionary=targets_labeled_instance,
            ref_point=self.get_ref_point(),
            scale=self.get_scale(),
            axis=self.axis,
            label_fluorophore=label_fluorophore,
            plotting_params=plotting_params,
        )
        return instance_constructor

    def generate_instance(self):
        """
        Generate the instance by constructing emitters and reference points based on loaded source and label information.
        """
        if self.sequential_labelling:
            print("Sequential_labelling")
            primaries = self._generate_instance_constructor()
            n_epitopes = primaries["n_epitopes"]
            self._generate_primary_epitopes(**primaries)
            secondaries = self._generate_secondary_labelling()
            self.add_emitters_n_refpoint(n_epitopes= n_epitopes, **secondaries)
            # return primaries
        elif self.status["source"] and self.status["labels"]:
            constructor = self._generate_instance_constructor()
            self.add_emitters_n_refpoint(**constructor)
        else:
            print("Missing source or Label info")

    def add_structural_integrity(
        self,
        integrity=1,
        xmer_neigh_distance=100,
        fracture=-24,
        eps1=20,
        minsamples=1,
    ):
        """
        Specify parameters for adding structural_integrity to the instance.

        Parameters
        ----------
        deg_dissasembly : float, optional
            Degree of disassembly. Default is 0.5.
        xmer_neigh_distance : float, optional
            Neighbour distance for clustering. Default is 100.
        fracture : float, optional
            Fracture parameter. Default is -24.
        eps1 : float, optional
            Clustering parameter. Default is 20.
        minsamples : int, optional
            Minimum samples for clustering. Default is 1.
        """
        if integrity <= 1 and integrity >= 0:
            deg_dissasembly = 1 - integrity
        else:
            deg_dissasembly = 0
        if deg_dissasembly == 0:
            self.structural_integrity = False
            self.structural_integrity_target_normals = None
            self.generate_instance()
        else:
            d_cluster_params = dict(
                eps1=eps1,
                minsamples1=minsamples,
                eps2=xmer_neigh_distance,
                minsamples2=minsamples,
            )
            self.structural_integrity_params["d_cluster_params"] = d_cluster_params
            self.structural_integrity_params["deg_dissasembly"] = deg_dissasembly
            self.structural_integrity_params["xmer_neigh_distance"] = xmer_neigh_distance
            self.structural_integrity_params["fracture"] = fracture
            self.structural_integrity = True
            self.generate_instance()

    def _model_structural_integrity(self, target_normals_dictionary):
        """
        target_normals_dictionary: dictionary with keys: "coordinates"
        and "normals".
        valye of key Normals can be none.
        Returns:
            target_nomral_subset (dict):
                coordinates: numpy array of epitope coordinates
                normals: numpy array of unit vector to define nomral
                    from its corresponding epitope
        """
        # print(f"input for _model_structural_integrity: {target_normals_dictionary}")
        target_sites = target_normals_dictionary["coordinates"]
        # print(f"in model_structural_integrity: {target_sites.shape}")
        # print(target_sites, self.structural_integrity_params)
        boolean_subset = xmersubset_byclustering(
            epitopes_coords=target_sites,
            return_ids=True,
            **self.structural_integrity_params,
        )
        coor = target_normals_dictionary["coordinates"][boolean_subset,]
        if target_normals_dictionary["normals"] is not None:
            nor = target_normals_dictionary["normals"][boolean_subset,]
        else:
            nor = None
        target_nomral_subset = dict(coordinates=coor, normals=nor)
        target_nomral_subset

        return target_nomral_subset

    # methods to create instance
    def _add_emitters(self, emitters_dictionary: dict):
        self.emitters = emitters_dictionary

    def _set_ref_point(self, newref_point: np.array):
        self.params["ref_point"] = newref_point

    def _set_scale(self, newscale: float):
        self.params["scale"] = newscale

    def _set_axis(self, axis: dict):
        self.axis = axis

    def _set_label_fluorophopre(self, label_fluo: dict):
        self.label2fluo = label_fluo

    def add_emitters_n_refpoint(
        self,
        emitters_dictionary: dict,
        ref_point: np.ndarray,
        scale: float,
        axis: dict,
        label_fluorophore: dict,
        plotting_params: dict,
        n_epitopes = None,
        **kwargs,
    ):
        """
        Add emitters and reference point information to the instance.

        Parameters
        ----------
        emitters_dictionary : dict
            Dictionary of emitters per label.
        ref_point : numpy.ndarray
            Reference point.
        scale : float
            Scale of the coordinates.
        axis : dict
            Axis definition.
        label_fluorophore : dict
            Mapping from label to fluorophore.
        plotting_params : dict
            Plotting parameters for each label.
        **kwargs
            Additional keyword arguments.
        """
        self.labelnames = list(emitters_dictionary.keys())
        self._add_emitters(
            dict(emitters_dictionary)
        )  # making an explicit copy
        self._set_ref_point(ref_point)
        self._set_scale(scale)
        self._set_axis(dict(axis))
        self._set_label_fluorophopre(dict(label_fluorophore))
        self._gen_fluo2labels()
        self.current_n_epitopes = n_epitopes
        self.plotting_params = dict(plotting_params)

    def get_axis(self):
        """
        Get the axis information for the instance.

        Returns
        -------
        dict
            Axis definition with 'pivot' and 'direction'.
        """
        return self.axis

    def get_ref_point(self):
        """
        Get the reference point of the instance.

        Returns
        -------
        numpy.ndarray
            Reference point coordinates.
        """
        return self.params["ref_point"]

    def reorient_axis_euler(self, phi=0, theta=0, psi=0, order = "zyx", reset_orientation=False, **kwargs):
        """
        Rotations are extrinsic.

        
        theta: rotation around x axis in degrees
        phi: rotation around z axis in degrees
        psi: rotation around y axis in degrees
        """
        combined_R = R.from_euler(order, [phi,theta, psi], degrees=True)
        new_vector = combined_R.apply(self.axis["direction"])
        self.transform_reorient_axis(neworientation=new_vector, reset_orientation=reset_orientation)
    
    def reorient_axis_by_plane(self, xy=0, yz=0, xz=0, reset_orientation=False, sequential=False, **kwargs):
        """
        Relative rotations from the current particle axis. 
        Since we use euler rotations to find the new vector,
        an xy-plane rotation does not equal a rotation aroun the axis when it points to the z direction

        """
        if sequential:
            self.reorient_axis_euler(theta=yz, reset_orientation=reset_orientation)
            self.reorient_axis_euler(psi=xz, reset_orientation=reset_orientation)
            self.reorient_axis_euler(phi=xy, reset_orientation=reset_orientation)
        else: 
            if yz != 0:
                self.reorient_axis_euler(theta=yz, reset_orientation=reset_orientation)
            elif xz != 0:
                self.reorient_axis_euler(psi=xz, reset_orientation=reset_orientation)
            elif xy != 0:
                self.reorient_axis_euler(phi=xy, reset_orientation=reset_orientation)


    def reset_axis_orientation(self):
        self.transform_reorient_axis(neworientation=self.axis_reset["direction"])
        

    def transform_reorient_axis(self, neworientation: np.array, reset_orientation=False):
        """
        Reorient the instance to a new orientation.

        Parameters
        ----------
        neworientation : numpy.ndarray
            New orientation vector.
        """
        if reset_orientation:
            self.reset_axis_orientation()
        else:
            if np.linalg.norm(neworientation) == 0:
                print(
                    f"Norm for vector {neworientation} is 0. No reorientation done"
                )
            else:
                thet = np.arccos(
                    np.dot(self.axis["direction"], neworientation)
                    / (
                        np.linalg.norm(self.axis["direction"])
                        * np.linalg.norm(neworientation)
                    )
                )
                if np.absolute(thet) == 1:
                    print(
                        f"input vector {neworientation} "
                        f'has same direction as {self.axis["direction"]}. '
                        f"No reorientation done"
                    )
                else:
                    nori = copy.copy(neworientation)
                    # this function should take the new orientation and match the
                    for trgt in self.labelnames:
                        if self.get_emitter_by_target(trgt) is not None:
                            reoriented = rotate_pts_by_vector(
                                self.get_emitter_by_target(trgt),
                                self.axis["direction"],
                                nori,
                                self.get_ref_point(),
                            )
                            self.emitters[trgt] = reoriented
                    self.axis["direction"] = nori

    def transform_rotate_around_axis(self, degree: float = 0.0):
        theta_radians = degree * (math.pi / 180)


        for labeltype in self.emitters.keys():
            if self.get_emitter_by_target(labeltype) is not None:
                self.emitters[labeltype] = planar_rotation(
                    points=self.get_emitter_by_target(labeltype), 
                    points_reference=self.get_ref_point(), 
                    unitary_vector_of_rotation=self.axis["direction"], 
                    theta=theta_radians)
                # rotate sources as well
                self.source["targets"][labeltype]["coordinates"] = planar_rotation(
                    points=self._get_source_coords_normals(labeltype)[
                            "coordinates"
                        ], 
                    points_reference=self.get_ref_point(), 
                    unitary_vector_of_rotation=self.axis["direction"], 
                    theta=theta_radians)

    def transform_translate(self, newcenter):
        """
        Translate the instance to a new center.

        Parameters
        ----------
        newcenter : numpy.ndarray
            New center coordinates.
        """
        nref = copy.copy(newcenter)
        # move labels
        for labeltype in self.emitters.keys():
            if self.get_emitter_by_target(labeltype) is not None:
                self.emitters[labeltype], _ = transform_displace_set(
                    self.get_emitter_by_target(labeltype),
                    self.get_ref_point(),
                    nref,
                )
                # translate sources as well
                self.source["targets"][labeltype]["coordinates"], _ = (
                    transform_displace_set(
                        self._get_source_coords_normals(labeltype)[
                            "coordinates"
                        ],
                        self.get_ref_point(),
                        nref,
                    )
                )


        # then replace reference point
        self._set_ref_point(nref)
        self.axis["pivot"] = nref

    def get_info(self):
        print(
            f"instance has {len(self.emitters.keys())} "
            f"types of emitters with names {list(self.emitters.keys())}"
        )

    def get_scale(self):
        """
        Get the current scale of the instance.

        Returns
        -------
        float
            Scale value.
        """
        return self.params["scale"]

    def scale_coordinates_system(self, new_scale: float):
        scaling_factor = self.params["scale"] / new_scale
        # scaling_factor_source = self.source["scale"] / new_scale
        # scale object data
        self.params["ref_point"] = self.params["ref_point"] * scaling_factor
        self.axis["pivot"] = self.axis["pivot"] * scaling_factor
        self.radial_hindance *= scaling_factor
        self._set_scale(new_scale)
        # scale source data
        # self.source["reference_pt"] *= scaling_factor_source
        # self.source["axis"]['pivot'] *= scaling_factor_source
        # self.source["scale"] = new_scale
        # scale emitters and each target in source
        for labeltype in self.emitters.keys():
            # print("scaling")
            if self.get_emitter_by_target(labeltype) is not None:
                # print(f'before: {self.emitters[labeltype]}')
                self.emitters[labeltype] = (
                    self.get_emitter_by_target(labeltype) * scaling_factor
                )
            self.source["targets"][labeltype]["coordinates"] *= scaling_factor
            if self.source["targets"][labeltype]["normals"] is not None:
                print(f"normals not NONE: {scaling_factor}")
                print(
                    self.source["targets"][labeltype]["normals"],
                )
                self.source["targets"][labeltype]["normals"] *= scaling_factor
        # scale probes data
        for labeltype in self.labels.keys():
            probe_scaling_factor = self.labels[labeltype]["scale"] / new_scale
            if self.labels[labeltype]["emitters"] is not None:
                self.labels[labeltype]["emitters"] = (
                    self.labels[labeltype]["emitters"].astype("float64")
                    * probe_scaling_factor
                )
            if self.labels[labeltype]["minimal_distance"] is not None:
                self.labels[labeltype][
                    "minimal_distance"
                ] *= probe_scaling_factor
            if "coordinates" in self.labels[labeltype].keys():
                self.labels[labeltype]["coordinates"] = (
                    self.labels[labeltype]["coordinates"].astype("float64")
                    * probe_scaling_factor
                )
            if (
                self.labels[labeltype]["binding"]["distance"]["to_target"]
                is not None
            ):
                self.labels[labeltype]["binding"]["distance"][
                    "to_target"
                ] *= probe_scaling_factor
            if (
                self.labels[labeltype]["binding"]["distance"][
                    "between_targets"
                ]
                is not None
            ):
                self.labels[labeltype]["binding"]["distance"][
                    "between_targets"
                ] *= probe_scaling_factor
            self.labels[labeltype]["scale"] = new_scale
            # update the primary targets if they exist
            if labeltype in self.primary["targets"].keys():
                self.primary["targets"][labeltype][
                    "coordinates"
                ] *= probe_scaling_factor
                if self.primary["targets"][labeltype]["normals"] is not None:
                    self.primary["targets"][labeltype][
                        "normals"
                    ] *= probe_scaling_factor
        for labeltype in self.secondary.keys():
            probe_scaling_factor = (
                self.secondary[labeltype]["scale"] / new_scale
            )
            if self.secondary[labeltype]["emitters"] is not None:
                self.secondary[labeltype]["emitters"] = (
                    self.secondary[labeltype]["emitters"].astype("float64")
                    * probe_scaling_factor
                )
            if self.secondary[labeltype]["minimal_distance"] is not None:
                self.secondary[labeltype][
                    "minimal_distance"
                ] *= probe_scaling_factor
            if "coordinates" in self.secondary[labeltype].keys():
                self.secondary[labeltype]["coordinates"] = (
                    self.secondary[labeltype]["coordinates"].astype("float64")
                    * probe_scaling_factor
                )
            if (
                self.secondary[labeltype]["binding"]["distance"]["to_target"]
                is not None
            ):
                self.secondary[labeltype]["binding"]["distance"][
                    "to_target"
                ] *= probe_scaling_factor
            if (
                self.secondary[labeltype]["binding"]["distance"][
                    "between_targets"
                ]
                is not None
            ):
                self.secondary[labeltype]["binding"]["distance"][
                    "between_targets"
                ] *= probe_scaling_factor
            self.secondary[labeltype]["scale"] = new_scale
        if self.structural_integrity:
            self.structural_integrity_params["d_cluster_params"]["eps1"] *= scaling_factor
            self.structural_integrity_params["d_cluster_params"]["eps2"] *= scaling_factor
            self.structural_integrity_params["xmer_neigh_distance"] *= scaling_factor

    # methods to get emitters by target name

    def get_emitter_by_target(self, targetname: str):
        """
        Get emitter coordinates for a given target.

        Parameters
        ----------
        targetname : str
            Name of the target.

        Returns
        -------
        numpy.ndarray or None
            Emitter coordinates or None if not found.
        """
        if self.emitters[targetname] is None:
            return None
        elif len(self.emitters[targetname]) == 0:
                return None
        else:
            return self.emitters[targetname]

    def _gen_fluo2labels(self):
        inv_map = {}
        for k, v in self.label2fluo.items():
            inv_map[v] = inv_map.get(v, []) + [k]
        self.fluo2labels = inv_map

    def get_emitters_by_fluorophore(self, fluoname: str):
        """
        Get all emitter coordinates for a given fluorophore.

        Parameters
        ----------
        fluoname : str
            Name of the fluorophore.

        Returns
        -------
        numpy.ndarray
            Array of emitter coordinates.
        """
        lab_target = self.fluo2labels[fluoname]
        pulled = np.empty((0, 3))
        for l in lab_target:
            if self.get_emitter_by_target(l) is not None:
                pulled = np.vstack([pulled, self.get_emitter_by_target(l)])
        return pulled

    # Visualisation methods
    def show_probe(
        self,
        probe_name=None,
        axesoff=False,
        return_plot=False,
        view_init=[30, 0, 0],
        xlims=[-100, 100],
        ylims=[-100, 100],
        zlims=None,
        central_axis=True,
        with_structural_atoms=False,
        emitter_plotsize=10,
        emitter_plotcolour="#984ea3",
        emitter_plotalpha=1,
        atoms_fraction=0.1,
        atoms_plotsize=1,
        atoms_plotalpha=1,
        use_dol=False,
        **kwargs,
    ):
        """
        Visualize the probe structure in 3D.

        Parameters
        ----------
        probe_name : str, optional
            Name of the probe to visualize.
        axesoff : bool, optional
            If True, hide axes. Default is False.
        return_plot : bool, optional
            If True, return the matplotlib axis object. Default is False.
        view_init : list of int, optional
            Initial view angles [elev, azim, roll]. Default is [30, 0, 0].
        xlims : list, optional
            X-axis limits. Default is [-100, 100].
        ylims : list, optional
            Y-axis limits. Default is [-100, 100].
        zlims : list, optional
            Z-axis limits. Default is None.
        **kwargs
            Additional keyword arguments.

        Returns
        -------
        matplotlib.axes._subplots.Axes3DSubplot or None
            The axis if return_plot is True, otherwise None.
        """
        if probe_name is None:
            probe_name = list(self.labels.keys())[0]
            print(probe_name)
        fig = plt.figure()
        ax = fig.add_subplot(111, projection="3d")
        # probe_plotting_params = self._get_label_plotting_params(probe_name)
        total_coordinates = copy.copy(self.labels[probe_name]["emitters"])
        total_number_coordinates = total_coordinates.shape[0]
        center = np.mean(total_coordinates, axis=0)
        dol = self.labels[probe_name]["conjugation_sites"]["DoL"]
        int_dol = np.random.poisson(lam=dol)
        probe_emitters = total_coordinates[2:, :].reshape(
                total_number_coordinates - 2, 3
            )
        centered_emitters, translation_vector1 = transform_displace_set(
            probe_emitters, center, np.array([0, 0, 0])
        )
        if dol is not None and use_dol:
            list_reoriented_points = []
            int_dol = np.random.poisson(lam=dol)
            max_emitters = centered_emitters.shape[0]
            if int_dol != 0:
                if int_dol > max_emitters:
                    # use max value
                    pass
                else:
                    emitter_indices = np.arange(max_emitters)
                    selected_emitters = np.random.choice(
                        emitter_indices, size=int_dol, replace=False
                    )
                    for se in selected_emitters:
                        list_reoriented_points.append(centered_emitters[se])
                    centered_emitters = np.array(list_reoriented_points)
        probe_axis = total_coordinates[0:2, :]
        
        centered_axis, translation_vector2 = transform_displace_set(
            probe_axis, center, np.array([0, 0, 0])
        )
        ##
        add_ax_scatter(
            ax,
            format_coordinates(
                centered_emitters, plotmarker="o", plotsize=emitter_plotsize,
                plotalpha=emitter_plotalpha, plotcolour=emitter_plotcolour,
                
            ),
        )
        if with_structural_atoms:
            subset_atoms = sample_array(
                array=self.labels[probe_name]["structural_atoms"],
                fraction=atoms_fraction)
            centered_atoms, translation_vector1 = transform_displace_set(
                    subset_atoms, center, np.array([0, 0, 0])
                )
            add_ax_scatter(
                ax,
                format_coordinates(
                    centered_atoms, plotmarker="o", plotsize=atoms_plotsize,
                    plotalpha=atoms_plotalpha,
                ),
            )
            
        if central_axis:
            axis_segment = dict()
            axis_segment["pivot"] = centered_axis[0]
            axis_segment["direction"] = centered_axis[1] - centered_axis[0]
            draw1nomral_segment(
                axis_segment, ax, lenght=100, colors=["g", "y"]
            )
        ax.set_xlim(xlims)
        ax.set_ylim(ylims)
        if zlims is None:
            zlims = [
                np.min(centered_axis[:, 2]),
                np.min(centered_emitters[:, 2]),
            ]
            ax.set_zlim(zlims)
        else:
            ax.set_zlim(zlims)
        ax.view_init(elev=view_init[0], azim=view_init[1], roll=view_init[2])
        ax.set_box_aspect(
            [
                ub - lb
                for lb, ub in (getattr(ax, f"get_{a}lim")() for a in "xyz")
            ]
        )
        if axesoff:
            ax.set_axis_off()
        else:
            ax.set_xlabel("X (Å)")
            ax.set_ylabel("Y (Å)")
            ax.set_zlabel("Z (Å)")
        if return_plot:
            return ax
        else:
            fig.show()

    def show_instance(
        self,
        labelnames="All",
        reference_point=False,
        view_init=[20, 0, 0],
        axesoff=True,
        show_axis=False,
        with_sources=False,
        return_plot=False,
        source_size=None,
        emitter_plotsize=None,
    ):
        """
        Visualize the labelled instance in 3D.

        Parameters
        ----------
        labelnames : str or list, optional
            Label names to visualize. Default is "All".
        reference_point : bool, optional
            If True, show the reference point. Default is False.
        view_init : list of int, optional
            Initial view angles [elev, azim, roll]. Default is [20, 0, 0].
        axesoff : bool, optional
            If True, hide axes. Default is True.
        show_axis : bool, optional
            If True, show the axis. Default is False.
        with_sources : bool, optional
            If True, show source coordinates. Default is False.
        return_plot : bool, optional
            If True, return the matplotlib axis object. Default is False.
        source_size : int, optional
            Size of source markers. Default is 1.
        emitter_plotsize : int, optional
            Size of emitter markers. Default is 1.

        Returns
        -------
        matplotlib.axes._subplots.Axes3DSubplot or None
            The axis if return_plot is True, otherwise None.
        """
        if labelnames == "All":
            fig = plt.figure()
            ax = fig.add_subplot(111, projection="3d")
            for labs in self.labelnames:
                lab_plotparams = self._get_label_plotting_params(labs)
                if emitter_plotsize is not None:
                    lab_plotparams["plotsize"] = emitter_plotsize
                add_ax_scatter(
                    ax,
                    format_coordinates(self.emitters[labs], **lab_plotparams),
                )
                if with_sources:
                    if source_size is not None:
                        source_size = source_size
                    else:
                        source_size = 1
                    add_ax_scatter(
                        ax,
                        format_coordinates(
                            self._get_source_coords_normals(labs)[
                                "coordinates"
                            ],
                            plotsize=source_size,
                        ),
                    )
                # add_ax_scatter(ax, format_coordinates(self.emitters[labs]))
        if reference_point:
            ref = self.get_ref_point()
            print(f"Reference point at {ref}")
            ax.scatter(
                ref[0], ref[1], ref[2], c="k", label="ref", s=20, marker="x"
            )
        if show_axis:
            print(f'current axis direction: {self.axis["direction"]}')
            draw1nomral_segment(self.axis, ax, lenght=150, colors=["g", "y"])
        ax.view_init(elev=view_init[0], azim=view_init[1], roll=view_init[2])
        ax.set_box_aspect(
            [
                ub - lb
                for lb, ub in (getattr(ax, f"get_{a}lim")() for a in "xyz")
            ]
        )
        if axesoff:
            ax.set_axis_off()
        else:
            if self.get_scale() == 1e-10:
                ax.set_xlabel("X (Angstroms)")
                ax.set_ylabel("Y (Angstroms)")
                ax.set_zlabel("Z (Angstroms)")
            elif self.get_scale() == 1e-9:
                ax.set_xlabel("X (Nanometers)")
                ax.set_ylabel("Y (Nanometers)")
                ax.set_zlabel("Z (Nanometers)")
        if return_plot:
            return ax
        else:
            fig.show()

    def gen_axis_plot(
        self,
        labelnames="All",
        reference_point=False,
        view_init=[20, 0, 0],
        axesoff=True,
        show_axis=False,
        with_sources=False,
        source_plotsize=None,
        axis_object=None,
        emitter_plotsize=None,
        emitter_plotmarker="o",
        emitter_plotcolour="#984ea3",
        source_plotcolour="#01579D",
        source_plotmarker="D",
        source_plotalpha=1,
        with_normals=False,
        **kwargs,
    ):
        """
        Generate a 3D axis plot for the labelled instance.

        Parameters
        ----------
        labelnames : str or list, optional
            Label names to visualize. Default is "All".
        reference_point : bool, optional
            If True, show the reference point. Default is False.
        view_init : list of int, optional
            Initial view angles [elev, azim, roll]. Default is [20, 0, 0].
        axesoff : bool, optional
            If True, hide axes. Default is True.
        show_axis : bool, optional
            If True, show the axis. Default is False.
        with_sources : bool, optional
            If True, show source coordinates. Default is False.
        source_plotsize : int, optional
            Size of source markers. Default is 1.
        axis_object : matplotlib.axes._subplots.Axes3DSubplot, optional
            Axis object to plot on.
        emitter_plotsize : int, optional
            Size of emitter markers. Default is 1.
        """
        zlims = [0, 1]
        if labelnames == "All":
            for labs in self.labelnames:
                lab_plotparams = self._get_label_plotting_params(labs)
                if emitter_plotsize is not None:
                    lab_plotparams["plotsize"] = emitter_plotsize
                if emitter_plotmarker is not None:
                    lab_plotparams["plotmarker"] = emitter_plotmarker
                if emitter_plotcolour is not None:
                    lab_plotparams["plotcolour"] = emitter_plotcolour
                if self.emitters[labs].shape[0] == 0:
                    print(f'No emitters for label {labs}')
                else:
                    add_ax_scatter(
                        axis_object,
                        format_coordinates(self.emitters[labs], **lab_plotparams),
                    )
                    zlims[0] = np.min(self.emitters[labs])
                    zlims[1] = np.max(self.emitters[labs])
                if with_sources:
                    if source_plotsize is not None:
                        source_plotsize = source_plotsize
                    else:
                        source_plotsize = 1
                    if self.structural_integrity_target_normals is not None:
                        #add_ax_scatter(
                        #    axis_object,
                        #    format_coordinates(
                        #        self.structural_integrity_target_normals["coordinates"]
                        #    ),
                        #)
                        add_ax_scatter(
                            axis_object,
                            format_coordinates(
                                self._get_source_coords_normals(labs)[
                                    "coordinates"
                                ],
                                plotcolour=source_plotcolour,
                                plotsize=source_plotsize,
                                plotmarker=source_plotmarker,
                                plotalpha=source_plotalpha
                            ),
                        )
                    else:
                        add_ax_scatter(
                            axis_object,
                            format_coordinates(
                                self._get_source_coords_normals(labs)[
                                    "coordinates"
                                ],
                                plotcolour=source_plotcolour,
                                plotsize=source_plotsize,
                                plotmarker=source_plotmarker,
                                plotalpha=source_plotalpha
                            ),
                        )
                    if with_normals:
                        draw_nomral_segments(
                            [
                                self.source["targets"][labs]["normals"],
                                self.source["targets"][labs]["coordinates"],
                            ],
                            axis_object,
                            colors=["grey", "green"],
                            **kwargs,
                        )

                # add_ax_scatter(ax, format_coordinates(self.emitters[labs]))
        if reference_point:
            ref = self.get_ref_point()
            print(f"Reference point at {ref}")
            axis_object.scatter(
                ref[0], ref[1], ref[2], c="k", label="ref", s=20, marker="x"
            )
        if show_axis:
            print(f'current axis direction: {self.axis["direction"]}')
            draw1nomral_segment(
                self.axis, axis_object, lenght=150, colors=["g", "y"]
            )
        axis_object.view_init(
            elev=view_init[0], azim=view_init[1], roll=view_init[2]
        )
        if axesoff:
            axis_object.set_axis_off()
        else:
            fontsize = 5
            axis_object.set_zlim(zlims[0], zlims[1])
            # axis_object.set_zticks(np.arange(-200, 200, 200))
            axis_object.set_xlabel("X (Angstroms)", size=fontsize)
            axis_object.set_ylabel("Y (Angstroms)", size=fontsize)
            axis_object.set_zlabel("Z (Angstroms)", size=fontsize)
            axis_object.tick_params(
                axis="both", which="major", labelsize=fontsize
            )
        axis_object.set_box_aspect(
            [
                ub - lb
                for lb, ub in (
                    getattr(axis_object, f"get_{a}lim")() for a in "xyz"
                )
            ]
        )


def create_particle(source_builder=None, label_params_list=None):
    """
    Create a LabeledInstance particle from a source builder and label parameters.

    Parameters
    ----------
    source_builder : dict, optional
        Dictionary with source parameters.
    label_params_list : list of dict, optional
        List of label parameter dictionaries.

    Returns
    -------
    LabeledInstance
        The created particle instance.
    """
    particle = LabeledInstance()
    if source_builder is not None:
        particle.load_source(**source_builder)  # define the source
    if label_params_list is not None:
        particle = add_label_params_to_particle(particle, label_params_list)
    particle.generate_instance()
    return particle


def add_label_params_to_particle(particle: LabeledInstance, label_params):
    """
    Add label parameters to an initialized particle containing target sites.

    Parameters
    ----------
    particle : LabeledInstance
        The particle to add label parameters to.
    label_params : list of dict
        List of label parameter dictionaries.

    Returns
    -------
    LabeledInstance
        The particle with added label parameters.
    """
    targets = dict()
    for lab in label_params:
        # print(lab)
        fluorophore = lab["fluorophore"]
        coordinates = None
        secondary = False
        if lab["target"]["type"] == "Primary":
            print(f'Label: {lab["target"]["value"]} is secondary antibody')
            lab["label_name"] = lab["target"]["value"]
            secondary = True
        if "coordinates" in lab.keys():
            # print(lab["coordinates"])
            coordinates = np.array(lab["coordinates"])
            # print(coordinates, type(coordinates))
        targets[fluorophore] = coordinates
        particle.load_label(
            targets=targets,
            secondary=secondary,
            minimal_distance=lab["binding"]["distance"]["between_targets"],
            **lab,
        )
        particle.sequential_labelling = secondary
        if "epitope_site" in lab.keys():
            print(f'assigning epitope site : {lab["epitope_site"]}')
            particle.source["targets"]["epitope_site"] = copy.copy(
                lab["epitope_site"]
            )
    return particle



from scipy.signal import convolve, convolve2d
import numpy as np
from tqdm import tqdm
import math
import skimage
from scipy.ndimage import zoom
import copy


# keep
def convolve3D(voxelated_field, psf, method="fft"):
    """
    psf is the 3D superpixelated psf.
    The pixelsize of both psf and voxelated field must be equal
    psf is also assumed to be centered
    """
    result = convolve(voxelated_field, psf, mode="same", method=method)
    return result


"""
Section: convolve in 3D, use zrange to get emitter that are to be considered
"""


# keep
def voxelate_points_withrange(coordinates, bin_pixelsize, ranges):
    """
    ideally, this calculation is done in nm scale
    all variables are assumed to have the same dimensions
    pixelsize refers to the detection
    imsize, in pixels, are the number of pixel per dimensions
    bin_pixelsize is the voxel size of the data after binning
    """

    binrange = [ranges[0], ranges[1], ranges[2]]
    nbins = [
        int(np.diff(ranges[0])[0] / bin_pixelsize),
        int(np.diff(ranges[1])[0] / bin_pixelsize),
        int(np.diff(ranges[2])[0] / bin_pixelsize),
    ]
    hist3D, edges = np.histogramdd(coordinates, bins=nbins, range=binrange)
    return hist3D


# keep
def voxelize_active_fluorophores_withrange(
    coordinates, photons_in_frame, ranges, bin_pixelsize
):
    """
    - photons_in_frame: is a vector with correspondance 1-1 with the emitters
        coordinates and contains the amount of photons emited in the time frame
    - Coordinates: is a numpy array of shape Nx3, being N the number of
        emitters
    - range: is a list of 3 elemtns, one per dimension (XYZ in that order),
        each entry is a tuple of 2 elements stating
        the minimum and maximum absolute value that will be used to create
        the discrete field
    """
    activated = photons_in_frame > 0
    activated_coordinates = coordinates[activated, :]
    photons_per_emitter = photons_in_frame[activated]
    coordinates_intensity = []
    coordinates_intensity = np.repeat(
        activated_coordinates, repeats=photons_per_emitter, axis=0
    )
    voxel = voxelate_points_withrange(coordinates_intensity, bin_pixelsize, ranges)
    return voxel


## keep
def frame_by_volume_convolution(
    coordinates,
    photon_vector,
    ranges,
    bin_size,
    kernel3D,
    zfocus_slice,
    projection_depth=2,
    activation_only=False,
    asframe=True,
    as_mask=False,
):
    """
    Generate a 2D image from the  3D field of active emitters
    convolved with the 3D PSF
    The resulting 2D image can be generated by considering the whole
    volume or a subset from it.
    projection_depth tells the depth of the volume to be considered
    in units of frames

    IMPORTANT: All units are in the same scale, except image dimensions
    """
    projection_depth_half = int(projection_depth / 2)
    intensity_voxel = voxelize_active_fluorophores_withrange(
        coordinates, photon_vector, ranges, bin_pixelsize=bin_size
    )
    if as_mask:
        frame = np.sum(np.array(intensity_voxel),axis=2)
        return frame
    if activation_only:
        return intensity_voxel
    if np.sum(photon_vector) > 0:
        convolved_intensity = convolve3D(intensity_voxel, kernel3D)
    else:
        # there's no need to calculate the convolution if there are no photons
        convolved_intensity = intensity_voxel
    subset_of_frames = (
        zfocus_slice - projection_depth_half,
        zfocus_slice + projection_depth_half,
    )
    if projection_depth_half > int(convolved_intensity.shape[2] / 2):
        projection_depth_half = int(convolved_intensity.shape[2] / 2)
    if asframe:
        frame = np.sum(
            np.array(
                convolved_intensity[
                    :, :, zfocus_slice - projection_depth_half : zfocus_slice + projection_depth_half
                ]
            ),
            axis=2,
        )
        return frame
    else:
        return convolved_intensity
    

# keep
def generate_frames_volume_convolution(
    field_scale: float,
    field_coordinates: np.array,
    field_size: tuple,
    field_pixelsizeXY: float,
    field_zfocus: float,
    photons_frames: np.array,
    psf_scale: float,
    psf_array: np.array,
    psf_zstep: float,
    psf_focus_slice: int,
    psf_pixelsizeXY: float,
    asframes=True,
    activation_only=False,
    psf_projection_depth=1,
    as_mask=False,
    **kwargs,
):
    """
    Inputs are taken from the method _homogenise_scales4convolution_modality
    """
    if asframes:
        cframes = []
        nframes = np.shape(photons_frames)[1]
        for f in tqdm(range(int(nframes))):
            photons_in_frame_vect = photons_frames[
                :, f
            ]  # extract the vector that corresponds to the current frame
            # calculate depth. This value considers the PSF dimenstions
            coordinates_scaled = field_coordinates
            pixelsize_scaled = int(field_pixelsizeXY)
            zplane = field_zfocus
            depth = (
                psf_array.shape[2] / 2
            ) * psf_zstep  # Assumes the focus plane is at the middle
            ranges = [
                (0, field_size[0] * pixelsize_scaled),
                (0, field_size[1] * pixelsize_scaled),
                (zplane - depth, zplane + depth),
            ]
            frame_n = frame_by_volume_convolution(
                coordinates_scaled,
                photons_in_frame_vect,
                ranges,
                psf_zstep,
                psf_array,
                psf_focus_slice,
                projection_depth=psf_projection_depth,
                as_mask=as_mask
            )
            cframes.append(frame_n)

        stack_zxy = np.array(cframes)
        currentpixelsizes = (psf_pixelsizeXY, psf_pixelsizeXY)
        newpixelsizes = (int(field_pixelsizeXY), int(field_pixelsizeXY))
        binned_sequence = lateral_binning_stack(
            stack_zxy, currentpixelsizes, newpixelsizes, field_size
        )
        return binned_sequence
    else:
        volume_frames = []
        nframes = np.shape(photons_frames)[1]
        for f in tqdm(range(int(nframes))):
            photons_in_frame_vect = photons_frames[
                :, f
            ]  # extract the vector that corresponds to the current frame
            # calculate depth. This value considers the PSF dimenstions
            coordinates_scaled = field_coordinates
            pixelsize_scaled = int(field_pixelsizeXY)
            zplane = field_zfocus
            depth = (
                psf_array.shape[2] / 2
            ) * psf_zstep  # Assumes the focus plane is at the middle
            ranges = [
                (0, field_size[0] * pixelsize_scaled),
                (0, field_size[1] * pixelsize_scaled),
                (zplane - depth, zplane + depth),
            ]
            v_frame = frame_by_volume_convolution(
                coordinates_scaled,
                photons_in_frame_vect,
                ranges,
                psf_zstep,
                psf_array,
                psf_focus_slice,
                projection_depth=psf_projection_depth,
                asframe=asframes,
                activation_only=activation_only
            )
            volume_frames.append(v_frame)
        return volume_frames


# keep
def lateral_binning_stack(stack, currentpixelsizes, newpixelsizes, field_size):
    """
    stack is assumed to have the shape ZXY

    field_size is a touple that contains the absolute size of each dimension
    """
    print("Binning image stack")
    nframes = stack.shape[0]  ## Z in in the dimension 0
    binned_images = np.zeros((nframes, field_size[0], field_size[0]))
    blocks = np.floor_divide(
        newpixelsizes,
        currentpixelsizes,
    )
    for i in tqdm(range(int(nframes))):
        binned_images[i] = skimage.measure.block_reduce(
            stack[i],
            block_size=(int(blocks[0]), int(blocks[1])),
            func=np.sum,
            func_kwargs={"dtype": np.float32},
        )
    return binned_images


# keep
def discretise_field_xyz(coordinates: np.ndarray, binsizes: list, ranges: list):
    """
    Generate a 3D histogram from a set of coordinates.

    Parameters:
    - coordinates (np.ndarray): The coordinates to be discretised.
    - binsizes (list): The size of the bins in each dimension.
    - ranges (list): The range of values in each dimension.

    Returns:
    - hist3D (np.ndarray): The 3D histogram of the discretised field.
    - edges (list): The bin edges for each dimension.
    """
    binrange = [ranges[0], ranges[1], ranges[2]]
    nbins = [
        int(np.diff(ranges[0]) / binsizes[0]),
        int(np.diff(ranges[1]) / binsizes[1]),
        math.ceil(np.diff(ranges[2]) / binsizes[2]),
    ]
    hist3D, edges = np.histogramdd(coordinates, bins=nbins, range=binrange)
    return hist3D, edges


"""
Convolution functions as 2D slices. Convolve plane-wise with 2D PSF.
Methods for optimising lookup convolution
"""


# keep
def _prepare_data_4convolutions(
    field_coordinates: np.array,
    field_size: tuple,
    field_pixelsizeXY: float,
    field_zfocus: float,
    photons_frames: np.array,
    psf_array: np.array,
    psf_zstep: float,
    psf_focus_slice: int,
    psf_pixelsizeXY: float,
    **kwargs,
):
    psf_half_range = (
        min((psf_array.shape[2] - psf_focus_slice), psf_focus_slice) * psf_zstep
    )
    zrange = psf_half_range * 2
    nframes = np.shape(photons_frames)[1]
    n_emitters = np.shape(photons_frames)[0]
    # pixel dimensions for binning
    currentpixelsizes = (psf_pixelsizeXY, psf_pixelsizeXY)
    newpixelsizes = (int(field_pixelsizeXY), int(field_pixelsizeXY))
    # limits to discretise
    ranges_xyz = [
        (0, field_size[0] * field_pixelsizeXY),
        (0, field_size[1] * field_pixelsizeXY),
        (field_zfocus - zrange / 2, field_zfocus + zrange / 2),
    ]
    # #
    bin_pixelsizes_xyz = [psf_pixelsizeXY, psf_pixelsizeXY, psf_zstep]
    data4conv = dict(
        field_coordinates=field_coordinates,
        field_size=field_size,
        nframes=nframes,
        n_emitters=n_emitters,
        bining_lateral_pixelsize=psf_pixelsizeXY,
        zrange=zrange,
        currentpixelsizes=currentpixelsizes,  # used at binning
        newpixelsizes=newpixelsizes,  # used at binning
        field_zfocus=field_zfocus,
        ranges_xyz=ranges_xyz,
        bin_pixelsizes_xyz=bin_pixelsizes_xyz,
        photons_frames=photons_frames,
        psf_array=psf_array,
        psf_focus_slice=psf_focus_slice,
    )
    return data4conv


# keep
def generate_frames_projection_conv_optimised2(
    field_scale: float,
    field_coordinates: np.array,
    field_size: tuple,
    field_pixelsizeXY: float,
    field_zfocus: float,
    photons_frames: np.array,
    psf_scale: float,
    psf_array: np.array,
    psf_zstep: float,
    psf_focus_slice: int,
    psf_pixelsizeXY: float,
    **kwargs,
):
    conv_data = _prepare_data_4convolutions(
        field_coordinates,
        field_size,
        field_pixelsizeXY,
        field_zfocus,
        photons_frames,
        psf_array,
        psf_zstep,
        psf_focus_slice,
        psf_pixelsizeXY,
    )

    convolved_emitters_list = create_convolved_emitters_list(**conv_data)
    #  Then construct every frame by adding their corresponding emitters
    frames = construct_frames(convolved_emitters_list, **conv_data)
    binned_frames = bin_timeseries(frames, **conv_data)
    return np.array(convolved_emitters_list), frames, binned_frames


# keep
def construct_frames(convolved_emitters_list, nframes, photons_frames, **kwargs):
    print("constructing frames")
    timeserie = []
    for f in tqdm(range(int(nframes))):
        # extract the vector that corresponds to the current frame
        photons_in_frame_vect = photons_frames[:, f]
        timeserie.append(
            merge_emissions_lookup(photons_in_frame_vect, convolved_emitters_list)
        )
    return np.array(timeserie)


# keep
def bin_timeseries(timeseires, currentpixelsizes, newpixelsizes, field_size, **kwargs):
    # field_size is equivalent to image size
    # assumes the stack has dimensions tXY
    print(
        f"current pixels size for bining: {currentpixelsizes}; new ones {newpixelsizes}"
    )
    print("Binning images")
    binned_timeserie = lateral_binning_stack(
        timeseires, currentpixelsizes, newpixelsizes, field_size
    )
    return binned_timeserie


# keep
def create_convolved_emitters_list(
    field_coordinates,
    n_emitters,
    psf_array,
    ranges_xyz,
    bin_pixelsizes_xyz,
    psf_focus_slice,
    field_zfocus,
    **kwargs,
):
    """ """
    print("Calculating single-emitter profile")
    convolved_emitters_list = []
    for em in tqdm(range(n_emitters)):
        emitter_coordinate = field_coordinates[em, :]
        row_vector_coordinate = emitter_coordinate[
            np.newaxis, :
        ]  # to preserve dimension of 1X3
        convolved_emitters_list.append(
            single_emitter_convolved_lookup(
                row_vector_coordinate,
                ranges_xyz,
                bin_pixelsizes_xyz,
                psf_array,
                psf_focus_slice,
                field_zfocus,
            )
        )
    return convolved_emitters_list


# keep
def single_emitter_convolved_lookup(
    coordinate,
    ranges_xyz,
    bin_pixelsizes_xyz,
    psf_array,
    psf_focus_slice,
    field_zfocus,
    **kwargs,
):
    """
    coordinate has to have shape (1,3)

    """
    discrete_emitter_volume, edges = discretise_field_xyz(
        coordinate, bin_pixelsizes_xyz, ranges_xyz
    )
    # wchich slice contains the emitter
    where_ = np.where(discrete_emitter_volume > 0)
    slice_of_emiter = where_[2][0]
    # which slice on the discrete version corresponds to the focus
    where_discrete_focus = np.where(np.isclose(edges[2], field_zfocus))
    discrete_focus_idx = where_discrete_focus[0][0]
    # according to this get which would be the corresponding slice of PSF
    psf_slice_id = psf_focus_slice - (discrete_focus_idx - slice_of_emiter)
    discrete_2D_slice = discrete_emitter_volume[:, :, slice_of_emiter]
    psf_slice2D = psf_array[:, :, psf_slice_id]
    # convolve these 2D images
    convolved_emitter = convolve2d(discrete_2D_slice, psf_slice2D, mode="same")
    # Profit
    return convolved_emitter


# keep
def merge_emissions_lookup(photons_in_frame_vect, convolved_emitters_list):
    """
    This method will take the pre-calculated emittsion profile
    from each isolated emitter in the field
    and according to the photon vector will combine the
    emission from the active emitters and its corresponding
    photon flux at that frame
    """
    # exception if there are no activated emitters in frame
    total_photons_frame = np.sum(photons_in_frame_vect)
    frame = np.zeros(convolved_emitters_list[0].shape)
    if total_photons_frame > 0:
        for pos in range(len(photons_in_frame_vect)):
            if photons_in_frame_vect[pos] > 0:
                intensity = convolved_emitters_list[pos] * photons_in_frame_vect[pos]
                frame = np.sum([frame, intensity], axis=0)
    # shall return a single image
    return frame



from datetime import datetime
import numpy as np
import os

def write_txt(list_of_lines: list, dirname: str, notes: str):
    now = datetime.now()  # dd/mm/YY H:M:S
    sim_dt_string = now.strftime("%Y%m%d") + "_"
    filename = dirname + sim_dt_string + notes + ".csv"
    current_files = os.listdir(dirname)
    if filename.split("/")[-1] in current_files:
        count = 1
        new_filename = (
            dirname
            + sim_dt_string
            + notes
            + "_"
            + str(count)
            + ".csv"
        )
        while new_filename.split("/")[-1] in current_files:
            count += 1
            new_filename = (
                dirname
                + sim_dt_string
                + notes
                + "_"
                + str(count)
                + ".csv"
            )
        filename = new_filename
    text_file = open(filename, "w")
    for li in range(len(list_of_lines)):
        text_file.write(list_of_lines[li])
        text_file.write("\n")
    text_file.close()
    print(f"csv written: {filename}")



import vlab4mic.generate.labelled_instance as lp
from vlab4mic import workflows
from vlab4mic.utils import data_format
import pytest
import copy
import numpy as np


def test_empty_particle():
    particle = lp.LabeledInstance()

# test with source and label params from fixture

structure_list = [
    "7R5K",
]
labels_per_structure = {
    "7R5K": [
        "NPC_Nup96_Cterminal_direct",
    ]
}
generic_labels = [
    "NHS_ester",
]


def test_add_probe_with_peptide_motif(experiment_7r5k_base):
    test_experiment = copy.copy(experiment_7r5k_base)
    test_experiment.remove_probes()
    list_of_proteins = test_experiment.structure.list_protein_names()
    chain_name = list_of_proteins[np.random.randint(0, len(list_of_proteins))]
    probe_name = "Antibody"
    test_experiment.add_probe(
        probe_template=probe_name,
        peptide_motif={
            "chain_name": chain_name,
            "position": "cterminal",
        },
    )
    assert len(test_experiment.probe_parameters) == 1
    assert (
        test_experiment.probe_parameters[probe_name]["label_name"]
        == probe_name
    )
    assert (
        test_experiment.probe_parameters[probe_name]["target"]["type"]
        == "Sequence"
    )
    assert (
        type(test_experiment.probe_parameters[probe_name]["target"]["value"])
        == str
    )



import numpy as np
from .points_transforms import decorate_epitopes_normals
from ..sample.arrays import binomial_epitope_sampling
from ..transform.normals import add_wobble


def do_sym_operation(pts_set, sym_operation):
    # pts_set contains the coordinates of al epitopes/atoms of an asymetric unit
    # sym is a list of 2 objects, the first is a matrix that establishes a rotation
    rot = sym_operation[0]
    translation = sym_operation[1]
    symmetry_transformed = []
    for epi in pts_set:
        rotated = np.matmul(rot, epi)  ## matrix * vector
        rot_trans = rotated + translation
        symmetry_transformed.append(rot_trans)
    pts_set_symm = np.array(symmetry_transformed)
    return pts_set_symm


def sym_transforms_numeric_only(
    assembly_transf_dictionary,
    point_coords,
    by_sym_unit=False,
    show_transforms_used=False,
):
    # point coords is a numpy array with dimensions nx3,
    # where n is the number of points to transform
    # assembly_transf_dictionary is a dicitonary
    # where each key is the name of the transform,
    # and the value of each key a list
    # the first element of the list is the rotation matrix,
    # the second element is the translation vector
    new_points = []
    for op in assembly_transf_dictionary.keys():
        # ignore transformations that have letters
        flag = True
        try:
            # try converting to integer
            int(op)
        except ValueError:
            flag = False
        if show_transforms_used:
            print(op, flag)
        if flag:
            new_points.append(
                do_sym_operation(point_coords, assembly_transf_dictionary[op])
            )
    new_points_per_sym = np.array(new_points)
    if by_sym_unit:
        return new_points_per_sym
    else:
        allatoms = new_points_per_sym.reshape(
            (np.shape(new_points_per_sym)[0]) * (np.shape(new_points_per_sym)[1]), 3
        )
        return allatoms


def summarize_epitope_atoms(atoms_per_epitope, method="average", residue_pos=0):
    # this function takes the output list from get_epitopes_by_sequence()
    # Each element of the list is the colection of atoms that correspond to that epitope
    nepitopes = len(atoms_per_epitope)
    epitopes_summary = np.zeros((nepitopes, 3))
    if method == "average":
        for i in range(nepitopes):
            epitopes_summary[i, :] = np.mean(atoms_per_epitope[i], axis=0)
    if method == "last":
        for i in range(nepitopes):
            epitopes_summary[i, :] = atoms_per_epitope[i][nepitopes - 1,]
    if method == "first":
        for i in range(nepitopes):
            epitopes_summary[i, :] = atoms_per_epitope[i][0,]
    if method == "residue":
        pass
    return epitopes_summary


def array_coords_subset(
    array, percentage
):  # Note: this method is deprecated and will be removed in future versions. 
    # Instead, use "sample_array" method found at vlab4mic.utils.sample.sample_array
    # show a percentage of all the atoms contained in the full assembly of a CIF
    # this assumes the shape of array is N x 3, where N is the total number of atoms
    if percentage < 1:
        size = array.shape[0]
        ids = np.random.choice(
            np.arange(0, size), int(size * percentage), replace=False
        )
        subset = array[ids, :]
        return subset
    elif percentage == 1:
        return array


def create_instance_label(
    coords_normals: dict, target_type: str, label_data: dict, **kwargs
):
    # receive points + normals for the target
    # receive the labeling parameters
    """
    This function is intended to be called for each type of label
    in the source of labelling instance

    coords_normals: dictionary as follows
        "coordinates" = numpy.array
        "normals" = None or numpy array
        ...
    """
    if target_type == "direct":
        # it could be that the corresponding label is indirect
        # but would not be able to compute since there are no normals
        return direct_labelling(coords_normals, label_data)
    elif target_type == "indirect":
        return indirect_labelling(coords_normals, label_data)


def direct_labelling(coords_normals, label_data, **kwargs):
    coordinates = coords_normals["coordinates"]
    efficiency = label_data["labelling_efficiency"]
    label_data["minimal_distance"]
    return binomial_epitope_sampling(
        epitopes=coordinates,
        p=efficiency,
        normals=None,
        min_distance=label_data["minimal_distance"]
    )


def indirect_labelling(coords_nomrals, label_data, **kwargs):
    """
    coords_nomrals is a dictionary containing both coordinates and
    the normals asociated with each
    """
    # label data contains emitter coordinates, labeling efficiency and more parameters
    # needed to create variable or static labeling

    # step1: verify that the target type and its label type correspond
    #       otherwise rise and exemption and create a direct labelling
    # step2: if they match sample the available epitopes sterically by
    #       considering the distance imposed by the label
    # step3: then, populate each of those sampled epitopes with the label
    #       coordinates following the orientation
    if label_data["emitters"] is not None:
        # first, if there is a size constraint, sample the maximum number
        # that will fit these constraint

        # second: from those, we take the labeling efficiency into account
        # this is assuming the labeling efficiency is a reflection of
        # the binding energy needed not the steric impediiment of nearby fluorophroes

        # select randomly, and according to labelling_efficiency
        # the places that will be labelled
        #efficiency = label_data["labelling_efficiency"]
        if label_data["conjugation_sites"]["DoL"] is not None:
            if label_data["conjugation_sites"]["DoL"] > 0:
                dol=label_data["conjugation_sites"]["DoL"]
            else:
                dol = None
        else:
            dol = None
        epitopes, n_epitopes, normals = binomial_epitope_sampling(
            epitopes=coords_nomrals["coordinates"],
            p=label_data["labelling_efficiency"],
            normals=coords_nomrals["normals"],
            min_distance=label_data["minimal_distance"]
        )  #####################
        # print(f"number of epitopes: {epitopes.shape}")
        if epitopes.shape[0] == 0:
            indirect_realisation = None
            list_reoriented_points_normals = None
        else:
            # add wobble to normals
            #normals_ft_epitopes = [normals, epitopes]
            if label_data["binding"]["wobble_range"]["theta"]:
                cone_angle = label_data["binding"]["wobble_range"]["theta"]
                print(f"Wobbling normals: {cone_angle}")
                
                w_normals = []
                origin = np.array([0, 0, 0])
                for i in range(normals.shape[0]):
                    n_wobbled = add_wobble(origin, normals[i,:], cone_angle_deg=cone_angle)
                    w_normals.append(n_wobbled)
                normals_ft_epitopes = [np.array(w_normals), epitopes]
            else:
                normals_ft_epitopes = [normals, epitopes]
            indirect_realisation, list_reoriented_points_normals = decorate_epitopes_normals(
                normals_ft_epitopes, label_data["emitters"], dol=dol
            )
    else:
        print("No emitters defined in label. Using direct labelling")
        indirect_realisation, n_epitopes, _ = direct_labelling(coords_nomrals, label_data)
        list_reoriented_points_normals = None
    return indirect_realisation, n_epitopes, list_reoriented_points_normals



import numpy as np
import yaml
import copy
import os
from Bio.PDB import PDBParser, MMCIFParser, PPBuilder, CaPPBuilder
from Bio.PDB.MMCIF2Dict import MMCIF2Dict
import matplotlib.pyplot as plt
import random

from ..utils.transform import cif_builder  # Verified
from ..utils.visualisation.matplotlib_plots import (
    draw1nomral_segment,
    add_ax_scatter,
    draw_nomral_segments,
)  # verified

from .labels import Label

from ..utils.data_format.visualisation import (
    format_coordinates,
    set_colorplot,
)  # verified

from ..utils.data_format.structural_format import builder_format  # verified

from ..utils.transform.normals import normals_by_scaling, global_normal_direction  # verified
from ..utils.sample import arrays


class MolecularStructureParser:
    def __init__(self):
        """
        Initialize a MolecularStructureParser object with default attributes.
        """
        self.id = ""
        self.chain_builder = "CaPPBuilder"
        self.ch_builder_max_dist = 5
        self.ppgen = None
        self.CIFdictionary = None
        self.protein_names = None
        self.assembly_operations = None
        self.chains_dict = None
        self.assembly_refpt = None
        self.assembly_atoms = None
        self.num_assembly_atoms = None
        self.store_atoms_assembly = True
        self.assymetric_defined = None
        self.label_targets = dict()
        self.label_normals = None
        # label_targets is a dictionary contining coordinates and parameters
        # for each labelling used. Keys are the name of the label
        self.plotting_params = dict()
        self.scale = 1e-10  # in meters
        self.axis = dict(pivot=None, direction=None)
        self.normals_params = {}
        self.normals_params["mode"] = "scaling"
        self.normals_params["normal_vector"] = None

    def _clear_labels(self):
        """
        Clear all label targets and normals, resetting plotting parameters except for assembly atoms.
        """
        #print("labels cleared")
        self.label_targets = dict()
        self.label_normals = None
        temp_ = self.plotting_params["assemblyatoms"]
        self.plotting_params = dict()
        self.plotting_params["assemblyatoms"] = temp_

    def initialise_parsers(self, dictionary: dict):
        """
        Create molecular structure of the PDB/CIF based on its format and associated parameters.

        Parameters
        ----------
        dictionary : dict
            Dictionary containing file path, ID, format, and optionally title.
        """
        print("Parsing structure. This might take a few seconds...")
        self.source_file = dictionary["file"]
        self.id = dictionary["ID"]
        self.format = dictionary["format"]
        try:
            self.identifier = dictionary["title"]
        except:
            self.identifier = ""
        if self.format == "PDB":
            parser = PDBParser()
            self.struct = parser.get_structure(self.identifier, self.source_file)
        elif self.format == "CIF":
            parser = MMCIFParser()
            self.struct = parser.get_structure(self.identifier, self.source_file)
        else:
            print(f"{self.format} is not a valid format. Valid options are PDB or CIF.")
        if self.chain_builder == "PPBuilder":
            self.ppgen = PPBuilder(self.ch_builder_max_dist)
        elif self.chain_builder == "CaPPBuilder":
            self.ppgen = CaPPBuilder(self.ch_builder_max_dist)

    def __str__(self):
        return f"Structure object for {self.identifier} of {self.format} format"

    # BUILD STRUCTURE
    def build_structure(self):
        """
        Build the structure by generating chains, assembly operations, and reference point.
        """
        print(f"Building structure for: {self.id}: {self.identifier}...")
        if self.chains_dict is None:
            self.generate_chains_sequences()
        if self.assembly_operations is None:
            self.generate_assemmbly_operations()
        if self.assembly_refpt is None:
            self.generate_assembly_reference_point()

    def generate_MMCIF_dictionary(self):
        """
        Generate a dictionary with all fields parsed from the CIF file.
        """
        # generate the dictionary with all field parsed from CIF file
        self.CIFdictionary = MMCIF2Dict(self.source_file)
        self._gen_protein_names()

    def get_atoms_infile(self):
        """
        Obtain all atom coordinates written in the PDB/CIF file.

        Returns
        -------
        numpy.ndarray
            Array of atom coordinates (N, 3).
        """
        coordinates = []
        for chains in self.struct:
            for chain in chains:
                for residue in chain:
                    for atom in residue:
                        coordinates.append(atom.get_vector().get_array())
        # get array is a method for the class vector of biopython.
        # np.array(atom.get_vector()) does not return the coordinates from the vector
        return np.array(coordinates)

    def generate_chains_sequences(self):
        """
        Construct a dictionary with chain IDs as keys and peptide sequences as values.
        """
        seq_chains_dict = {}
        for chains in self.struct:
            for chain in chains:
                pps = self.ppgen.build_peptides(chain)
                sequence = ""
                for peptide in pps:
                    sequence += peptide.get_sequence()
                seq_chains_dict[chain.get_id()] = str(sequence)
        self.chains_dict = seq_chains_dict
        # nchains = len(list(seq_chains_dict.keys()))
        # print(f"file contains {nchains} chains")
    
    def _gen_protein_names(self):
        """
        Generate a dictionary of protein names and associated chain/strand information.
        """
        protein_name = dict()
        try:
            chain_name = self.CIFdictionary["_entity.pdbx_description"]
            chain_number = self.CIFdictionary["_entity.id"]
            strand_id = self.CIFdictionary["_entity_poly.pdbx_strand_id"]
            for  name, number, strands in zip(chain_name, chain_number, strand_id):
                protein_name[name] = {"number": number, "strand_id": strands}
        except:
            pass
        self.protein_names = protein_name
    
    def list_protein_names(self):
        """
        List all protein names in the structure.

        Returns
        -------
        list of str
            List of protein names.
        """
        if self.protein_names is not None:
            if len(list(self.protein_names.keys())) > 0:
                return list(self.protein_names.keys())
        else:
            return []

    def _random_substring(string, size=5):
        """
        Generate a random substring of a given size from a string.

        Parameters
        ----------
        string : str
            Input string.
        size : int, optional
            Length of the substring. Default is 5.

        Returns
        -------
        str
            Random substring.
        """
        if len(string) > size:
            start = random.randint(0, len(string) - (size + 1))
            end = start + size
            return string[start:end]

    def _sequence_substring(self, string, size=5, position="random"):
        """
        Extract a substring from a sequence at a specified position.

        Parameters
        ----------
        string : str
            Input sequence.
        size : int, optional
            Length of the substring. Default is 5.
        position : str, optional
            Position to extract from ("random", "cterminal", "nterminal").

        Returns
        -------
        str
            Extracted substring.
        """
        if position == "random":
            return self._random_substring(string, size)
        elif position == "cterminal":
            start = len(string) - (size + 1)
            return string[start:(len(string)-1)]
        elif position == "nterminal":
            end = size 
            return string[0:end]

    def get_peptide_motif(self, chain_name=None, chain_id=None, size = 5, position="cterminal"):
        """
        Get a peptide motif from a specified chain and position.

        Parameters
        ----------
        chain_name : str, optional
            Name of the chain.
        chain_id : str, optional
            Chain ID.
        size : int, optional
            Length of the motif. Default is 5.
        position : str, optional
            Position in the chain ("cterminal", "nterminal", "random").

        Returns
        -------
        tuple
            (chain_name, chain_id, position, motif)
        """
        if chain_name is None:
            chains_in_structure = list(self.protein_names.keys())
            chain_name = random.choice(chains_in_structure)
        if chain_id is None:
            chain_id = random.choice(self.protein_names[chain_name]["strand_id"].split(","))
        chain_sequence = self.chains_dict[chain_id]
        return chain_name, chain_id, position, self._sequence_substring(chain_sequence, size=size, position=position)

    def generate_assemmbly_operations(self):
        """
        Parse the rotation/translation operations needed to construct a molecular assembly from an asymmetric unit.
        """
        if self.format == "PDB":
            print("Assembly operations are only defined for CIF format files.")
            self.assymetric_defined = False
        else:
            if self.CIFdictionary is None:  # check if already created
                self.generate_MMCIF_dictionary()
            if "_database_2.database_id" in self.CIFdictionary.keys():
                if "AlphaFoldDB" in self.CIFdictionary["_database_2.database_id"]:
                    self.assymetric_defined = False
            try:
                # get to know how many transformations there are
                transformation_ids = self.CIFdictionary["_pdbx_struct_oper_list.id"]
                # each element of the list is pair of matrix-vector needed to transform data
                # iterate over each of the transformations
                assembly_transformations = []
                for tr in range(len(transformation_ids)):
                    row1 = [
                        self.CIFdictionary["_pdbx_struct_oper_list.matrix[1][1]"][tr],
                        self.CIFdictionary["_pdbx_struct_oper_list.matrix[1][2]"][tr],
                        self.CIFdictionary["_pdbx_struct_oper_list.matrix[1][3]"][tr],
                    ]

                    row2 = [
                        self.CIFdictionary["_pdbx_struct_oper_list.matrix[2][1]"][tr],
                        self.CIFdictionary["_pdbx_struct_oper_list.matrix[2][2]"][tr],
                        self.CIFdictionary["_pdbx_struct_oper_list.matrix[2][3]"][tr],
                    ]

                    row3 = [
                        self.CIFdictionary["_pdbx_struct_oper_list.matrix[3][1]"][tr],
                        self.CIFdictionary["_pdbx_struct_oper_list.matrix[3][2]"][tr],
                        self.CIFdictionary["_pdbx_struct_oper_list.matrix[3][3]"][tr],
                    ]
                    rotation_matrix = np.array([row1, row2, row3], dtype="float")

                    vector = np.array(
                        [
                            self.CIFdictionary["_pdbx_struct_oper_list.vector[1]"][tr],
                            self.CIFdictionary["_pdbx_struct_oper_list.vector[2]"][tr],
                            self.CIFdictionary["_pdbx_struct_oper_list.vector[3]"][tr],
                        ],
                        dtype="float",
                    )
                    assembly_transformations.append([rotation_matrix, vector])
                struct_oper_dictionary = dict(zip(transformation_ids, assembly_transformations))
                if len(transformation_ids) > 1:
                    self.assymetric_defined = True
                        # print(
                        #    "This model is defined with more than one symmetric transformation.
                        # Will consider the assembly as assymetric defined"
                        # )
                else:
                    # when no assembly unit exist, there is only 1 transform expected
                    self.assymetric_defined = False
                    # print("This model is defined with only one symmetric transformation")
                self.assembly_operations = struct_oper_dictionary
            except:
                self.assymetric_defined = False
                

    def generate_assembly_reference_point(self):
        """
        Generate a reference point for the whole structure by constructing a full assembly and averaging all atom coordinates.
        """
        # this function could use a small portion of the atoms defined only
        allatoms_defined = self.get_atoms_infile()  # Parse all atoms in file
        self.atoms_in_file = allatoms_defined
        if self.assembly_operations is None:
            self.generate_assemmbly_operations()
        if self.assymetric_defined:
            # print(
            #    "Strcture is defined by assymetric units.
            # Generating symmetry partners with the following transformations."
            # )
            # hardcoded: use only transforms with integers as id
            assemblyatoms = cif_builder.sym_transforms_numeric_only(
                self.assembly_operations, allatoms_defined, show_transforms_used=False
            )
            assembly_reference_point = np.mean(assemblyatoms, axis=0)
            if self.store_atoms_assembly:
                # print("Storing assembly atoms")
                self.assembly_refpt = assembly_reference_point
                self.assembly_atoms = assemblyatoms
                self.num_assembly_atoms = np.shape(assemblyatoms)[0]
                assembly_plotting_params = dict(
                    plotsize=20, plotalpha=0.01, plotmarker="o", plotcolour="#000000"
                )
                self.plotting_params["assemblyatoms"] = assembly_plotting_params
            else:
                self.assembly_refpt = assembly_reference_point
            # define axis
            # print("Defining axis from rotations")
            vect = self._central_axis_from_rotations()
            self.set_axis_with_vector(vect)
            self.axis_reset = copy.copy(self.axis)
        else:
            # print(
            #    "Strcture is not defined by assymetric units. Using only atoms in file"
            # )
            if self.store_atoms_assembly:
                # print("Storing assembly atoms")
                self.assembly_refpt = np.mean(allatoms_defined, axis=0)
                assembly_plotting_params = dict(
                    plotsize=20, plotalpha=0.01, plotmarker="o", plotcolour="#000000"
                )
                self.assembly_atoms = allatoms_defined
                self.num_assembly_atoms = np.shape(allatoms_defined)[0]
                self.plotting_params["assemblyatoms"] = assembly_plotting_params
            else:
                self.assembly_refpt = np.mean(allatoms_defined, axis=0)
            self.set_axis_with_vector()  # default [0,0,1]
            self.axis_reset = copy.copy(self.axis)

    def _get_assemblyatoms(self):
        return self.assembly_atoms["coordinates"]

    def set_axis_from_point(self, axis_defining_point):
        """
        Set the axis of the structure using a point.

        Parameters
        ----------
        axis_defining_point : numpy.ndarray
            Point to define the axis direction.
        """
        direction = axis_defining_point - self.assembly_refpt
        self.axis = dict(pivot=self.assembly_refpt, direction=direction)

    # central axis definition

    def set_axis_with_vector(self, vector: np.array = np.array([0, 0, 1])):
        """
        Set the axis of the structure using a direction vector.

        Parameters
        ----------
        vector : numpy.ndarray, optional
            Direction vector. Default is [0, 0, 1].
        """
        self.axis = dict(pivot=self.assembly_refpt, direction=vector)

    def _central_axis_from_rotations(self):
        """
        Calculate the central axis orientation from the assembly operations.

        Returns
        -------
        numpy.ndarray
            Central axis vector.
        """
        average = np.identity(3)
        keys = list(self.assembly_operations.keys())
        for i in range(1, len(self.assembly_operations)):
            average = np.matmul(average, self.assembly_operations[keys[i]][0])
        eigenvalues, eigenvectors = np.linalg.eig(average)
        central_axis = eigenvectors[
            :, np.argmax(np.isclose(eigenvalues, 1.0))
        ]  # central axis is the eigenvector that corresponds to the eigenvalue = 1
        if central_axis.dtype == "complex128":
            central_axis = np.array([0, 0, 1])
        return central_axis

    # Parsing atoms

    def _get_atom_res_chain(self, chainnames: list, resnames: list, atomnames: list):
        """
        Parse the structure and return atom coordinates defined by atom name, residue ID, and chain ID.

        Parameters
        ----------
        chainnames : list of str
            List of chain IDs.
        resnames : list of str
            List of residue names.
        atomnames : list of str
            List of atom names.

        Returns
        -------
        numpy.ndarray
            Array of atom coordinates.
        """
        myatoms = []
        for model in self.struct:
            for chain in model:
                if chain.id in chainnames:
                    for residue in chain:
                        # this conditional checks if the residue is
                        # defined in resname list
                        if residue.resname.strip() in resnames:
                            for atom in residue:
                                if atom.name.strip() in atomnames:
                                    myatoms.append(atom.get_coord())
        return np.array(myatoms)

    def _get_site_specific(
        self, chainnames: list, resnames: list, atomnames: list, position: int
    ):
        """
        Parse the structure and return atom coordinates for a specific site.

        Parameters
        ----------
        chainnames : list of str
            List of chain IDs.
        resnames : list of str
            List of residue names.
        atomnames : list of str
            List of atom names.
        position : int
            Position in the chain.

        Returns
        -------
        numpy.ndarray
            Array of atom coordinates for the specific site.
        """
        print(f"looking in: {chainnames}, {resnames}, {atomnames}, {position}")
        myatoms = []
        for model in self.struct:
            for chain in model:
                if chain.id in chainnames:
                    for residue in chain:
                        #pos = pos + 1
                        pos = residue.id[1]
                        # this conditional checks if the residue is
                        # defined in resname list
                        print(
                            f"resname {pos}: {residue.get_resname()}, "
                        )
                        if residue.resname.strip() in resnames and position == pos:
                            for atom in residue:
                                if atom.name.strip() in atomnames:
                                    myatoms.append(atom.get_coord())
                                    print(
                                        f"site speficic found: {pos}, "
                                        f"{residue.resname.strip()}, "
                                        f"{atom.name.strip()} "
                                    )
        print(np.array(myatoms))
        return np.array(myatoms)

    def get_atom_res_chains_assembly(
        self, residues: list, atoms: list, position: int = None, chains: list[str]=None,
    ):
        """
        Parse the structure and return atom coordinates for specified residues and atoms, including assembly partners if defined.

        Parameters
        ----------
        residues : list of str
            List of residue names.
        atoms : list of str
            List of atom names.
        position : int, optional
            Position in the chain.

        Returns
        -------
        numpy.ndarray
            Array of atom coordinates.
        """
        if self.chains_dict is None:
            self.generate_chains_sequences()  # Initialized as none
        if self.assembly_refpt is None:
            self.generate_assembly_reference_point()  # Initialized as none
        if self.assembly_operations is None:
            self.generate_assemmbly_operations()
        if chains is None:
            list_of_chains = list(self.chains_dict.keys())
        else:
            list_of_chains = chains
        # chains in structure on which to look for
        # print(f"Using chains {list_of_chains}")
        if position is None:
            atom_res_chain = self._get_atom_res_chain(list_of_chains, residues, atoms)
        else:
            atom_res_chain = self._get_site_specific(
                list_of_chains, residues, atoms, position
            )
        # atom_res_chain only considers a signle asymmetric unit
        # if the CIF defines an assembly, the asymmetric partners
        # or atom_res_chain are generated as well
        if self.assymetric_defined:
            # print("Generating symmetry partners for atoms parsed")
            atom_res_chain_assembly = cif_builder.sym_transforms_numeric_only(
                self.assembly_operations, atom_res_chain, show_transforms_used=False
            )
            return atom_res_chain_assembly
        else:
            return atom_res_chain

    def gen_targets_by_atoms(self,
                             label_name = "atoms", 
                             residues = None,
                             atoms = None,
                             position = None,
                             chains = None,
                             **kwargs):
        """
        Generate target locations for labelling defined as direct labeling.

        Parameters
        ----------
        label_name : str, optional
            Name for the label. Default is "atoms".
        residues : list of str, optional
            List of residue names.
        atoms : list of str, optional
            List of atom names.
        position : int, optional
            Position in the chain.
        **kwargs
            Additional keyword arguments (e.g., fluorophore, labeling_efficiency).
        """
        try:
            fluorophore = kwargs["fluorophore"]
        except:
            fluorophore = None
        try:
            labeling_efficiency = kwargs["labeling_efficiency"]
        except:
            labeling_efficiency = None
        colour = set_colorplot(self.plotting_params)
        parsed = self.get_atom_res_chains_assembly(
            residues,
            atoms,
            position,
            chains
        )
        self.label_targets[label_name] = self._wrapup_label_dictionary(
            parsed, labeling_efficiency, fluorophore
        )
        label_plotting_params = dict(
            plotsize=20, plotalpha=1, plotmarker="o", plotcolour=colour
        )
        self.plotting_params[label_name] = label_plotting_params

    # Parsing secuences

    def gen_targets_by_sequence(self, target_name, sequence, **kwargs):
        """
        Generate target locations for labelling defined as indirect labeling.

        Parameters
        ----------
        target_name : str
            Name for the target.
        sequence : str
            Epitope sequence to search for.
        **kwargs
            Additional keyword arguments (e.g., fluorophore, labeling_efficiency, method).
        """
        #  only the first appearance of the epitope in every chain is retrieved
        #  if a chain has more than one epitope, only the first one is returned
        target_seq = sequence
        method = "average"
        if "method" in kwargs.keys():
            method = kwargs["method"]
        label_name = target_name
        try:
            fluorophore = kwargs["fluorophore"]
        except:
            fluorophore = None
        try:
            labeling_efficiency = kwargs["labeling_efficiency"]
        except:
            labeling_efficiency = None
        print(f"Searching for sequence: {target_seq}")
        if self.chains_dict is None:
            self.generate_chains_sequences()
        if self.assembly_refpt is None:
            self.generate_assembly_reference_point()
        if self.assembly_operations is None:
            self.generate_assemmbly_operations()
        epitopes_list = []
        for chains in self.struct:
            for chain in chains:
                pps = self.ppgen.build_peptides(chain)
                sequence = ""
                for peptide in pps:
                    sequence += peptide.get_sequence()
                start = sequence.find(target_seq)
                end = start + len(target_seq)
                if start != -1:
                    coords_chain = []
                    for residue, i in zip(chain, range(len(sequence))):
                        if i >= start and i <= end:
                            for atom in residue:
                                coords_chain.append(atom.get_vector().get_array())
                    coords_chain = np.array(coords_chain)
                    epitopes_list.append(coords_chain)
        # print(len(epitopes_list))
        # print(f"Summarizing epitopes location with method {method}")
        summarised_epitopes = cif_builder.summarize_epitope_atoms(
            epitopes_list, method=method
        )
        colour = set_colorplot(self.plotting_params)
        if self.assymetric_defined:
            print("Generating symmetry partners for epitopes parsed")
            epitopes_locations_assembly = cif_builder.sym_transforms_numeric_only(
                self.assembly_operations,
                summarised_epitopes,
                show_transforms_used=False,
            )
            self.label_targets[label_name] = self._wrapup_label_dictionary(
                epitopes_locations_assembly, labeling_efficiency, fluorophore
            )
            label_plotting_params = dict(
                plotsize=20, plotalpha=1, plotmarker="o", plotcolour=colour
            )
            self.plotting_params[label_name] = label_plotting_params
        else:
            self.label_targets[label_name] = self._wrapup_label_dictionary(
                summarised_epitopes, labeling_efficiency, fluorophore
            )
            label_plotting_params = dict(
                plotsize=20, plotalpha=1, plotmarker="o", plotcolour=colour
            )
            self.plotting_params[label_name] = label_plotting_params

    def _wrapup_label_dictionary(
        self, coordinates, labeling_efficiency, fluorophore, normals=None
    ):
        label_dictionary = dict(
            coordinates=coordinates,  # this makes the difference
            labeling_efficiency=labeling_efficiency,
            fluorophore=fluorophore,
            normals=normals,
        )
        return label_dictionary

    # generate targets from label info
    def gen_Targets(self, target_name,  target_type, target_value, **kwargs):
        """
        Generate targets for labelling based on the type of target.

        Parameters
        ----------
        target_name : str
            Name for the target.
        target_type : str
            Type of target ("Atom_residue" or "Sequence").
        target_value : Any
            Value for the target (e.g., atom/residue info or sequence).
        **kwargs
            Additional keyword arguments.
        """
        # call the generator depending on the target type
        if target_type == "Atom_residue":
            #print(target_value)
            self.gen_targets_by_atoms(target_name, **target_value)
        elif target_type == "Sequence":
            self.gen_targets_by_sequence(
                target_name=target_name,
                sequence=target_value)
        else:
            print(f"Label type {target_type} is not a valid label")

    def get_target_coords(self, targetid: str):
        """
        Get coordinates for a specific target.

        Parameters
        ----------
        targetid : str
            Target identifier.

        Returns
        -------
        numpy.ndarray
            Coordinates of the target.
        """
        return self.label_targets[targetid]["coordinates"]

    def get_target_normals(self, targetid: str):
        """
        Get normals for a specific target.

        Parameters
        ----------
        targetid : str
            Target identifier.

        Returns
        -------
        numpy.ndarray
            Normals of the target.
        """
        return self.label_targets[targetid]["normals"]

    def get_target_colour(self, targetid: str):
        """
        Get plotting colour for a specific target.

        Parameters
        ----------
        targetid : str
            Target identifier.

        Returns
        -------
        str
            Colour code.
        """
        return self.plotting_params[targetid]["plotcolour"]

    def set_labefficiency(self, targetname: str, newefficiency: float):
        """
        Set the labelling efficiency for a target.

        Parameters
        ----------
        targetname : str
            Target identifier.
        newefficiency : float
            New labelling efficiency value.
        """
        self.label_targets[targetname]["labeling_efficiency"] = newefficiency

    # visualisation
    def show_target_labels(
        self,
        labelnames: str = None,
        with_assembly_atoms=False,
        assembly_fraction=0.01,
        reference_point=True,
        view_init=[0, 0, 0],
        axesoff=True,
        show_axis=True,
        with_normals=False,
        return_plot=False,
        axis_object=None,
        target_size = None,
        target_plotcolour = None,
        atoms_size = None,
        atoms_alpha = None,
    ):
        """
        Visualize target labels and optionally assembly atoms and reference point.

        Parameters
        ----------
        labelnames : str, optional
            Name of the label to show. If None, show all.
        with_assembly_atoms : bool, optional
            If True, show assembly atoms. Default is False.
        assembly_fraction : float, optional
            Fraction of assembly atoms to show. Default is 0.01.
        reference_point : bool, optional
            If True, show the reference point. Default is True.
        view_init : list of int, optional
            Initial view angles [elev, azim, roll]. Default is [0, 0, 0].
        axesoff : bool, optional
            If True, hide axes. Default is True.
        show_axis : bool, optional
            If True, show the axis. Default is True.
        with_normals : bool, optional
            If True, show normals. Default is False.
        return_plot : bool, optional
            If True, return the matplotlib figure. Default is False.

        Returns
        -------
        matplotlib.figure.Figure or None
            The figure if return_plot is True, otherwise None.
        """
        if axis_object is not None:
            ax = axis_object
        else:
            fig = plt.figure()
            ax = fig.add_subplot(111, projection="3d")
        if labelnames is None:
            for trgt in list(self.label_targets.keys()):
                if target_size is not None:
                    self.plotting_params[trgt]["plotsize"] = target_size
                if target_plotcolour is not None:
                    self.plotting_params[trgt]["plotcolour"] = target_plotcolour
                if with_normals:
                    draw_nomral_segments(
                        [self.get_target_normals(trgt), self.get_target_coords(trgt)],
                        ax,
                        colors=["grey", self.get_target_colour(trgt)],
                    )
                else:
                    add_ax_scatter(
                        ax,
                        format_coordinates(
                            self.get_target_coords(trgt), **self.plotting_params[trgt]
                        ),
                    )
        if with_assembly_atoms:
            print(f"Showing {assembly_fraction*100}% of the total atoms")
            atoms_subset = arrays.sample_array(
                self.assembly_atoms, assembly_fraction
            )
            if atoms_size is not None:
                self.plotting_params["assemblyatoms"]["plotsize"] = atoms_size
            if atoms_alpha is not None:
                self.plotting_params["assemblyatoms"]["plotalpha"] = atoms_alpha
            add_ax_scatter(
                ax,
                format_coordinates(
                    atoms_subset, **self.plotting_params["assemblyatoms"]
                ),
            )
        if reference_point:
            print(f"Reference point at {self.assembly_refpt}")
            ax.scatter(
                self.assembly_refpt[0],
                self.assembly_refpt[1],
                self.assembly_refpt[2],
                c="k",
                label="Reoriented",
                s=20,
                marker="x",
            )
            self.assembly_refpt
        if show_axis:
            draw1nomral_segment(self.axis, ax, lenght=150, colors=["g", "y"])
        ax.view_init(elev=view_init[0], azim=view_init[1], roll=view_init[2])
        ax.set_box_aspect(
            [ub - lb for lb, ub in (getattr(ax, f"get_{a}lim")() for a in "xyz")]
        )
        if axesoff:
            ax.set_axis_off()
        if axis_object is not None:
            return ax
        else:
            if return_plot:
                plt.close()
                return fig
            else:
                fig.show

    def show_assembly_atoms(
        self, assembly_fraction=0.01, view_init=[0, 0, 0], axesoff=True, return_plot=False, show_axis=False
    ):
        """
        Visualize a fraction of the assembly atoms.

        Parameters
        ----------
        assembly_fraction : float, optional
            Fraction of assembly atoms to show. Default is 0.01.
        view_init : list of int, optional
            Initial view angles [elev, azim, roll]. Default is [0, 0, 0].
        axesoff : bool, optional
            If True, hide axes. Default is True.

        Returns
        -------
        matplotlib.figure.Figure
            The figure object.
        """
        print(f"Showing {assembly_fraction*100}% of the total atoms")
        fig = plt.figure()
        ax = fig.add_subplot(111, projection="3d")
        atoms_subset = arrays.sample_array(
            self.assembly_atoms, assembly_fraction
        )
        add_ax_scatter(
            ax,
            format_coordinates(atoms_subset, **self.plotting_params["assemblyatoms"]),
        )
        if show_axis:
            draw1nomral_segment(self.axis, ax, lenght=150, colors=["g", "y"])
        ax.view_init(elev=view_init[0], azim=view_init[1], roll=view_init[2])
        ax.set_box_aspect(
            [ub - lb for lb, ub in (getattr(ax, f"get_{a}lim")() for a in "xyz")]
        )
        if axesoff:
            ax.set_axis_off()
        if return_plot:
            plt.close()
            return fig
        else:
            fig.show()

    def create_instance_builder(self, write=False, savingdir=None):
        """
        Create a dictionary with information from target sites, their normals, axis, reference point, etc.
        This dictionary can be used as source for labelled structures in the simulation. 

        Parameters
        ----------
        write : bool, optional
            If True, write the builder to a YAML file. Default is False.
        savingdir : str, optional
            Directory to save the YAML file. If None, uses current working directory.

        Returns
        -------
        dict
            Instance builder dictionary.
        """
        targets = dict()
        for target_name, value in self.label_targets.items():
            targets[target_name] = dict(
                coordinates=value["coordinates"], normals=value["normals"]
            )
        instance_builder = builder_format(
            targets, self.assembly_refpt, self.scale, self.axis, self.identifier
        )
        if write:
            if savingdir is None:
                savingdir = os.getcwd()
            filename = savingdir + "/instance_builder_" + str(self.identifier) + ".yaml"
            with open(filename, "w") as file:
                print(f"Builder saved in: {filename}")
                yaml.dump(instance_builder, file)
        return instance_builder


class MolecularReplicates(MolecularStructureParser):
    def __init__(self, pdbxinfo):
        """
        Initialize a MolecularReplicates object from structure information.

        Parameters
        ----------
        pdbxinfo : dict
            Dictionary with structure file, title, format, and ID.
        """
        MolecularStructureParser.__init__(self)
        self.initialise_parsers(pdbxinfo)
        self.replicates = 1
        self.label_names = []
        self.plotcolours = {}
        self.label_fluorophore = {}

    def add_label(self, labelobj: Label):
        """
        Add a label to the molecular structure and generate its targets.

        Parameters
        ----------
        labelobj : Label
            Label object to add.
        """
        #prepare target value from label object
        self.gen_Targets(
            target_name=labelobj.get_name(),
            target_type=labelobj.get_target_type(),
            target_value=labelobj.params["target"]["value"]
            )
        self.label_names.append(labelobj.get_name())
        self.plotcolours[labelobj.get_name()] = labelobj.get_plotcolour()
        self.label_fluorophore[labelobj.get_name()] = labelobj.get_fluorophore()

    def assign_normals2targets(self, target: str = None):
        """
        Assign normals to targets using a specified method.

        Parameters
        ----------
        mode : str, optional
            Method for assigning normals ("scaling"). Default is "scaling".
        target : str, optional
            Specific target to assign normals to. If None, assign to all.
        """
        print(f"Assigning normals to targets with method: {self.normals_params['mode']}")
        if target is None:
            for target_name, value in self.label_targets.items():
                if self.normals_params["mode"] == "scaling":
                    normals = normals_by_scaling(
                        self.label_targets[target_name]["coordinates"]
                    )
                    self.label_targets[target_name]["normals"] = normals
                elif self.normals_params["mode"] == "global":
                    normals = global_normal_direction(
                        self.label_targets[target_name]["coordinates"],
                        normal_vector = self.normals_params["normal_vector"]
                    )
                    self.label_targets[target_name]["normals"] = normals
                elif self.normals_params["mode"] == "structure_axis":
                    normals = global_normal_direction(
                        self.label_targets[target_name]["coordinates"],
                        normal_vector = self.axis["direction"]
                    )
                    self.label_targets[target_name]["normals"] = normals
        else:
            if self.normals_params["mode"] == "scaling":
                normals = normals_by_scaling(self.label_targets[target]["coordinates"])
                self.label_targets[target]["normals"] = normals
            elif self.normals_params["mode"] == "global":
                normals = global_normal_direction(
                        self.label_targets[target_name]["coordinates"],
                        normal_vector = self.normals_params["normal_vector"]
                    )
                self.label_targets[target_name]["normals"] = normals
            elif self.normals_params["mode"] == "structure_axis":
                    normals = global_normal_direction(
                        self.label_targets[target_name]["coordinates"],
                        normal_vector = self.axis["direction"]
                    )
                    self.label_targets[target_name]["normals"] = normals


def build_structure_cif(
    cif_file: str, struct_title: str = "", cif_id: str = "", format_type="CIF"
):
    """
    Load and parse a PDB/CIF file and build a structure object.

    Parameters
    ----------
    cif_file : str
        Absolute path of the CIF file.
    struct_title : str, optional
        Title of the structure.
    cif_id : str, optional
        Structure ID.
    format_type : str, optional
        File format type ("CIF" or "PDB"). Default is "CIF".

    Returns
    -------
    MolecularReplicates
        Structure object with parsed atoms and information.
    """
    structure_dictionary = {
        "file": cif_file,
        "title": struct_title,
        "format": format_type,
        "ID": cif_id,
    }
    Molecularstructure = MolecularReplicates(structure_dictionary)
    Molecularstructure.build_structure()
    return Molecularstructure



import math
import numpy as np


def truncate(number, digits) -> float:
    # from Erwin Mayer at
    # https://stackoverflow.com/questions/8595973/truncate-to-three-decimals-in-python
    nbDecimals = len(str(number).split(".")[1])
    if nbDecimals <= digits:
        return number
    stepper = 10.0**digits
    return math.trunc(stepper * number) / stepper


def inCube(X, corners):
    """
    Check if a point `X` is inside of cube with `corners` (two points)
    # Based from: Michael Silverstein's answer
    at
    stackoverflow.com/questions/29720910/
    fastest-way-to-search-if-a-coordinate-is-inside-a-cube
    """
    # Where is X > corners?
    greater = X > corners
    # If X is greater than both corners of any axis, it is outside
    inside = ~np.any(np.equal(*greater))
    return inside


def generate_coordinate_textline(x, y, z, separator=",", decimals=3, **kwargs):
    line = (
        str(truncate(x, decimals))
        + separator
        + str(truncate(y, decimals))
        + separator
        + str(truncate(z, decimals))
    )
    return line


def dictionary2string(mydict):
    str_dict0 = str(mydict)
    str_dict1 = str_dict0.replace(" ", "")
    str_dict2 = str_dict1.replace(",", "_")
    str_dict3 = str_dict2.replace("'", "")
    str_dict4 = str_dict3.replace(":", "")
    str_dict5 = str_dict4.replace(":", "")
    return str_dict5[1:-1] + "_"



import numpy as np
import matplotlib.pyplot as plt
from datetime import datetime
import math
import copy
import yaml
from ..utils.io.yaml_functions import load_yaml

# from .psf.image_psf import load_psf
from ..utils.transform.datatype import *

from ..utils.transform import image_convolution as conv
from .psfs import elliptical_gaussian_3sigmas

from ..utils.io.text import write_txt
from ..utils.io.tiff import write_tif
from ..utils.data_format.visualisation import format_coordinates
from ..utils.visualisation.matplotlib_plots import add_ax_scatter
from ..utils.transform.noise import add_image_noise


class Imager:
    # Optical params
    def __init__(self):
        """
        Initialize the Imager object with default ROI and imaging parameters.
        """
        self.roi_params = {}
        self.roi_params["scale"] = 1e-6
        self.roi_params["reference_point"] = [0, 0]  # center of the imagexy
        self.roi_params["dimension_sizes"] = [1, 1, 1]  # size of the ROI
        self.roi_params["focus_plane"] = 0.0
        self.roi_params["plot_spacing"] = 0.1
        self.roi_params["ranges"] = []
        self.modalities = dict()

        self.field = {}
        self.field["reference_point"] = None
        self.fluorophore_params = (
            dict()
        )  # contain its photophysical parameters such as excitation and emission
        self.emitters_by_fluorophore = dict()
        self.emitters_by_channel = dict()
        self.writing_dir = ""
        self.identifier = ""

    def set_experiment_name(self, name: str):
        """
        Set the experiment identifier for the imager.

        Parameters
        ----------
        name : str
            Name or identifier for the experiment.
        """
        self.identifier = name

    def set_roi_from_file(self, roi_yaml: str):
        """
        Set ROI parameters from a YAML configuration file.

        Parameters
        ----------
        roi_yaml : str
            Path to the YAML file containing ROI parameters.
        """
        with open(roi_yaml, "r") as f:
            roi_p = yaml.safe_load(f)
        self.set_roi_params(**roi_p)

    # methods to get and set roi params
    def set_roi_params(self, **kwargs):
        """
        Set ROI parameters from keyword arguments.

        Parameters
        ----------
        **kwargs
            ROI parameter names and values.
        """
        if kwargs is not None:
            for key, value in kwargs.items():
                self.roi_params[key] = value
            self._calculate_roi_ranges()

    def _calculate_roi_ranges(self):
        dimensions = self.get_roi_params("dimension_sizes")
        reference_xy = self.get_roi_params("reference_point")
        referencez = self.get_roi_params("focus_plane")
        xrange = [
            (reference_xy[0] - (dimensions[0] / 2)),
            (reference_xy[0] + (dimensions[0] / 2)),
        ]
        yrange = [
            (reference_xy[1] - (dimensions[1] / 2)),
            (reference_xy[1] + (dimensions[1] / 2)),
        ]
        zrange = [
            (referencez - (dimensions[2] / 2)),
            (referencez + (dimensions[2] / 2)),
        ]
        self.roi_params["ranges"] = [xrange, yrange, zrange]

    def get_absolute_reference_point(self):
        """
        Get the absolute reference point in 3D (including focus plane).

        Returns
        -------
        numpy.ndarray
            Array of shape (1, 3) with the absolute reference point.
        """
        focus_plane = self.get_roi_params("focus_plane")
        ref_pt = self.get_roi_params("reference_point")
        ref_pt.append(focus_plane)
        # print(ref_pt)
        return np.array(ref_pt).reshape((1, 3))

    def set_focus(self, focus: float):
        """
        Set the focus plane position.

        Parameters
        ----------
        focus : float
            Z position of the focus plane.
        """
        self.roi_params["focus_plane"] = focus
        self._calculate_roi_ranges()

    def set_roi_position(self, x, y):
        """
        Set the (x, y) position of the ROI reference point.

        Parameters
        ----------
        x : float
            X coordinate.
        y : float
            Y coordinate.
        """
        self.roi_params["reference_point"] = [x, y]
        self._calculate_roi_ranges()

    def set_roi_sizes(self, x, y):
        """
        Set the size of the ROI in the x and y dimensions.

        Parameters
        ----------
        x : float
            Size in the x dimension.
        y : float
            Size in the y dimension.
        """
        self.roi_params["dimension_sizes"][0] = x
        self.roi_params["dimension_sizes"][1] = y
        self._calculate_roi_ranges()
        self._adjust_modalities_imsizes()

    def get_roi_params(self, parameter: str):
        """
        Get a copy of a specific ROI parameter.

        Parameters
        ----------
        parameter : str
            Name of the ROI parameter.

        Returns
        -------
        Any
            Value of the requested ROI parameter.
        """
        return copy.copy(self.roi_params[parameter])

    def import_field(
        self,
        field_emitters: dict,
        field_scale: float,
        plotting_params: dict,
        reference_point = None,
        field_sizes = None,
        particle_positions = None,
        particle_size = None,
        **kwargs,
    ):
        """
        Import emitter field and adjust ROI and emitter coordinates accordingly.

        Parameters
        ----------
        field_emitters : dict
            Dictionary of emitters by fluorophore.
        field_scale : float
            Scale of the field.
        plotting_params : dict
            Plotting parameters for emitters.
        reference_point : array-like
            Reference point of the field.
        field_sizes : array-like
            Size of the field in each dimension.
        **kwargs
            Additional keyword arguments.
        """
        if field_scale == self.get_roi_params("scale"):
            print("FOV scale and current scale are the same. No scaling performed")
            scaling_factor = 1
        else:
            scaling_factor = field_scale / self.get_roi_params("scale")
            # print(f"scaling factor set to {scaling_factor}")
        self.set_roi_sizes(
            field_sizes[0] * scaling_factor, field_sizes[1] * scaling_factor
        )
        self.field["reference_point"] = reference_point * scaling_factor
        self.recenter_roi()
        for fluoname, emitters in field_emitters.items():
            self.emitters_by_fluorophore[fluoname] = emitters * scaling_factor
        self.plotting_params = plotting_params
        self.particle_positions = particle_positions * scaling_factor
        if particle_size is not None:
            self.particle_size = particle_size * scaling_factor
        else: 
            self.particle_size = 0

    def recenter_roi(self):
        """
        Recenter the ROI based on the field's reference point.
        """
        field_xyz = self.field["reference_point"]
        # print(field_xyz)
        self.set_roi_position(field_xyz[0], field_xyz[1])
        self.set_focus(field_xyz[2])

    # fluorophore parameters
    def set_fluorophores_from_file(self, fluo_params: str):
        """
        Set fluorophore parameters from a YAML configuration file.

        Parameters
        ----------
        fluo_params : str
            Path to the YAML file containing fluorophore parameters.
        """
        fluo_p = load_yaml(fluo_params)
        self.set_fluorophores_params(**fluo_p)

    def set_fluorophores_params(
        self,
        identifier: str,
        photon_yield: int,
        emission: str,
        blinking_rates: dict,
        **kwargs,
    ):
        """
        Set parameters for a fluorophore.

        Parameters
        ----------
        identifier : str
            Name or ID of the fluorophore.
        photon_yield : int
            Number of photons emitted per second.
        emission : str
            Emission type.
        blinking_rates : dict
            Dictionary of blinking rate parameters.
        **kwargs
            Additional keyword arguments.
        """
        fluoname = identifier
        # print(fluo)
        if "plotcolour" in kwargs.keys():
            fluo_color = kwargs["plotcolour"]
        else:
            fluo_color = "blue"
        photons_per_second = photon_yield
        self.fluorophore_params[fluoname] = dict(
            photons_per_second=photons_per_second,
            emission=emission,
            blinking=dict(),
            plotcolour=fluo_color,
        )
        self.set_3state_blinking_params(fluoname, **blinking_rates)

    # self.set_3state_blinking_params()

    def set_fluorophore_photons_per_second(self, fluorophore_name, photon_yield):
        """
        Set the photon yield for a fluorophore.

        Parameters
        ----------
        fluorophore_name : str
            Name of the fluorophore.
        photon_yield : int
            Number of photons per second.
        """
        self.fluorophore_params[fluorophore_name]["photons_per_second"] = photon_yield

    def set_3state_blinking_params(
        self,
        fluorophore_name,
        kon,
        koff,
        kbleach,
        initial_state,
        photons_per_blink,
        **kwargs,
    ):
        """
        Set three-state blinking parameters for a fluorophore.

        Parameters
        ----------
        fluorophore_name : str
            Name of the fluorophore.
        kon : float
            On rate.
        koff : float
            Off rate.
        kbleach : float
            Bleaching rate.
        initial_state : int
            Initial state.
        photons_per_blink : float
            Photons emitted per blink.
        **kwargs
            Additional keyword arguments.
        """
        blink_dictionary = dict(
            kon=kon,
            koff=koff,
            kbleach=kbleach,
            initial_state=initial_state,
            photons_per_blink=photons_per_blink,
        )
        self.fluorophore_params[fluorophore_name]["blinking"] = dict(blink_dictionary)

    # modality module
    def set_imaging_modality_from_file(self, modality_file: str):
        """
        Set imaging modality parameters from a YAML configuration file.

        Parameters
        ----------
        modality_file : str
            Path to the YAML file containing modality parameters.
        """
        with open(modality_file, "r") as f:
            modality_p = yaml.safe_load(f)
        self.set_imaging_modality(**modality_p)

    def set_imaging_modality(
        self,
        filters: dict,
        psf_params: dict,
        detector: dict,
        emission="blinking",
        modality: str = "modality1",
        prints=False,
        **kwargs,
    ):
        """
        Set parameters for an imaging modality.

        Parameters
        ----------
        filters : dict
            Filter/channel configuration.
        psf_params : dict
            PSF parameters.
        detector : dict
            Detector parameters.
        emission : str, optional
            Emission type. Default is "blinking".
        modality : str, optional
            Name of the modality. Default is "modality1".
        prints : bool, optional
            If True, print debug information.
        **kwargs
            Additional keyword arguments.
        """
        self._create_modality(modality)
        self._set_modality_channels(modality, filters, prints=prints)
        self._set_modality_emission(modality, emission)
        self._set_modality_psf(modality, **psf_params, prints=prints)
        self._set_modality_detector(modality, **detector)

    def _create_modality(self, modality: str):
        self.modalities[modality] = dict(
            filters=None, detector=None, psf=None, emission=None
        )

    def _set_modality_channels(self, modality, fluorophores_in_channel, prints=False):
        if fluorophores_in_channel is None:
            if prints:
                print("Creating channel for each fluorophore")
            ch = 0
            filter_dictionary = dict()
            for fluoname in list(self.emitters_by_fluorophore.keys()):
                if prints:
                    print(fluoname)
                channel_name = "ch" + str(ch)
                fluorophores_in_channel = []
                fluorophores_in_channel.append(
                    fluoname
                )  # this is because a channel can have several fluorophores
                filter_dictionary[channel_name] = fluorophores_in_channel
                ch += 1
            self.modalities[modality]["filters"] = filter_dictionary
        else:
            self.modalities[modality]["filters"] = fluorophores_in_channel

    def _set_modality_psf(
        self, modality: str, stack_source: str = "generate", prints=False, **kwargs
    ):
        if stack_source == "generate":
            psf_stack, convolution_type = self._generate_analytical_PSF_stack(**kwargs, prints=prints)
            focus_plane = int((psf_stack.shape)[2] / 2)
            # print(f"focus plane: {focus_plane}")
        else:
            if prints:
                print(f"Loading PSF from file path: {stack_source}")
            psf_stack, convolution_type = self._load_PSF_from_file(
                stack_source, **kwargs
            )
            focus_plane = int(
                (psf_stack.shape)[2] / 2
            )  # this will be changed if the parameter is defined
        # initialising the psf parameters
        self.modalities[modality]["psf"] = dict(
            psf_stack=psf_stack,
            convolution_type=convolution_type,
            focus_plane=focus_plane,
        )
        for key, value in kwargs.items():
            self.modalities[modality]["psf"][key] = value
        # define the psf depth
        if "depth" not in self.modalities[modality]["psf"]:
            depth_default = int((psf_stack.shape)[2] / 2)
            if prints:
                print("No depth parameter found for psf, assigning default")
            self.modalities[modality]["psf"]["depth"] = depth_default

    def _set_modality_emission(self, modality, emission):
        self.modalities[modality]["emission"] = emission

    def _set_modality_detector(
        self,
        modality,
        image_size=[],  # deprecated, image size will only be taken from ROI
        pixelsize=None,
        bits_pixel=None,
        noise_model=None,
        noise_order=None,
        **kwargs,
    ):
        """
        Set noise and image parameters.
        """
        # image_size_ROI = self.get_roi_params("dimension_sizes")[0:2]
        modality_imsize = self._calculate_imsize_from_ROIranges(pixelsize)
        # print(f"image_size calculated: {modality_imsize} ")
        noise_model_0, noise_order_0 = self._gen_default_noise_model()
        self.modalities[modality]["detector"] = dict(
            image_size=[],
            pixelsize=pixelsize,
            bits_pixel=bits_pixel,
            noise_model=noise_model_0,
            noise_order=noise_order_0,
        )
        if noise_order is not None:
            self.set_noise_order(modality, noise_order)
        if noise_model is not None:
            for noise_t, params in noise_model.items():
                self.set_noise_model_param(modality, noise_t, params)
        self._set_image_size(modality, modality_imsize)

    def _calculate_imsize_from_ROIranges(
        self, mod_pixelsize
    ):  # There should be specified the scale of the pixelsize of modality
        roi_scale = self.get_roi_params("scale")
        xrange = self.get_roi_params("ranges")[0]
        yrange = self.get_roi_params("ranges")[1]
        xsize = int((xrange[1] - xrange[0]) / mod_pixelsize)
        ysize = int((yrange[1] - yrange[0]) / mod_pixelsize)
        imsize = [xsize, ysize]
        return imsize

    def _adjust_modalities_imsizes(self):
        for mod in self.modalities.keys():
            # print(f"adjusting imsize for {mod}")
            mode_pixelsize = self.modalities[mod]["detector"]["pixelsize"]
            modality_imsize = self._calculate_imsize_from_ROIranges(mode_pixelsize)
            # print(f"image_size calculated: {modality_imsize} ")
            self._set_image_size(mod, modality_imsize)

    def set_noise_order(self, modality, noise_order: list):
        """
        Set the order in which noise types are applied for a modality.

        Parameters
        ----------
        modality : str
            Name of the modality.
        noise_order : list of str
            List specifying the order of noise types.
        """
        self.modalities[modality]["detector"]["noise_order"] = noise_order
        # print(f"Noise order set as: {noise_order}")

    def set_noise_model_param(self, modality, noise_type, params: dict):
        """
        Set parameters for a specific noise type in a modality.

        Parameters
        ----------
        modality : str
            Name of the modality.
        noise_type : str
            Type of noise.
        params : dict
            Parameters for the noise type.
        """
        self.modalities[modality]["detector"]["noise_model"][noise_type] = params
        # print(f"Noise model set as: {noise_type} with params {params}")

    def _gen_default_noise_model(self):
        noise_model = dict(
            binomial={"p": 1},
            gamma={"g": 1},
            gaussian={"sigma": 1},
            poisson={"mock": None},
            conversion={"adu": 1},
            baselevel={"bl": 1},
        )
        noise_order = ["baselevel", "gaussian", "conversion"]
        return noise_model, noise_order

    def _set_image_size(self, modality, imsize):
        # adjust ROI size
        self.modalities[modality]["detector"]["image_size"] = imsize

    # generate and load PSFs
    def _generate_analytical_PSF_stack(
        self, shape=[24, 24, 24], std_devs=[1, 1, 1], prints=False, **kwargs
    ):
        """
        Generate a eliptical 3D Gaussian PSF
        Shape: in units of pixels
        std_devs: list of standard deviations per dimension
        """
        if prints:
            print(
                f"Generating unitary analytical PSF stack with shape {shape} "
                f"and standard deviations {std_devs}"
            )
        psf_stack = elliptical_gaussian_3sigmas(shape=shape, std_devs=std_devs)
        convolution_type = "3Dvolume"
        return psf_stack, convolution_type

    def _load_PSF_from_file(self, path: str, transpose, unitary, **kwargs):
        # print("loading PSF stack from file")
        psf_stack = load_psf(path, transpose, unitary)
        if unitary:
            convolution_type = "3Dvolume"
        else:
            # convolution_type = "2Dlookup"
            convolution_type = "2Dlookup_optim"
        return psf_stack, convolution_type

    # Imaging methods
    def generate_imaging(
        self,
        modality=None,
        channels: list = ["ch0",],
        nframes=1,
        nbeads=0,
        save=False,
        noise=False,
        exp_time=0.001,
        masks=False,
        **kwargs,
    ):
        """
        Master function that generates image sequences depending on the modality
        This function is responsible for
            Preparing the coordinates according to ROI (DONE!)
            Creating the Photons per frame matrix
            Take focus plane, PSF and photon budgets
            Feed these data into the Convolution method (or analytical)
            Adds noise
            Profit!
        """
        if modality is None:
            modality = list(self.modalities.keys())[0]
        # get fluorophores to simulate
        print(f"Simulating imaging from modality: {modality} in channel {channels}")
        output_per_channel = dict()
        output_per_channel_noiseless = dict()
        beads_per_channel = dict()
        beads_per_channel_noiseless = dict()
        for ch in channels:
            if ch not in self.modalities[modality]["filters"].keys():
                raise ValueError(
                    f"Channel {ch} not found in modality {modality} filters"
                )
            else:
                fluonames = list(self.modalities[modality]["filters"][ch])
                # prepare a dictionary that contains the emitters per channel defined
                output_per_fluoname = dict()
                writing_notes = self.identifier + "_" + str(modality) + "_" + str(ch) + "_"
                for fluo in fluonames:  # a channel could capture multiple fluorophores
                    # print(fluo)
                    writing_notes_fluo = writing_notes + str(fluo)
                    if masks:
                        emitters = self.get_emitters_in_ROI(fluo, masks=masks)
                    else:
                        emitters = self.get_emitters_in_ROI(fluo)
                    n_emitters = emitters.shape[0]
                    if n_emitters < 1:
                        no_emitters = True
                        # if no emitter is in range, an image with noise should be generated
                        emitters = np.array(
                            [
                                [0, 0, 0],
                            ]
                        )
                        photons_frames = np.repeat(0, nframes).reshape((1, nframes))
                        emission_notes = "None"
                        print(f"photons should be zero: {photons_frames}")
                        field_data, psf_data = self._homogenise_scales4convolution_modality(
                            modality, emitters, photons_frames
                        )

                    else:
                        no_emitters = False
                        # adjust the absolute XY positions to the ROI dimentions
                        emitters[:, 0] = emitters[:, 0] - self.roi_params["ranges"][0][0]
                        emitters[:, 1] = emitters[:, 1] - self.roi_params["ranges"][1][0]
                        photons_frames, emission_notes = self.calculate_photons_per_frame(
                            modality, fluo, n_emitters, nframes, exp_time
                        )
                        field_data, psf_data = self._homogenise_scales4convolution_modality(
                            modality, emitters, photons_frames
                        )
                    # write emitter positions after being placed in the FOV
                    gt_notes = writing_notes_fluo + "_usedForImaging"
                    emitters_to_export = field_data["field_coordinates"]
                    if save:
                        self.write_ground_truth_positions(
                            emitters_to_export, "x [nm],y [nm],z [nm]", gt_notes, no_emitters
                        )
                    if "convolution_type" in kwargs.keys():
                        convolution_type = kwargs["convolution_type"]
                    elif masks:
                        convolution_type = "mask"
                    else:
                        convolution_type = self.modalities[modality]["psf"]["convolution_type"]
                    psf_size = psf_data["psf_array"].shape
                    print(f"size of psf is: {psf_size}")
                    simparams = dict(
                        field_data=field_data,
                        psf_data=psf_data,
                        nbeads=nbeads,
                        photons_per_bead=self.fluorophore_params[fluo]["blinking"][
                            "photons_per_blink"
                        ],
                        psf_projection_depth=self.modalities[modality]["psf"]["depth"],
                    )
                    if convolution_type == "direct":
                        pass
                    elif convolution_type == "mask":
                        images = conv.generate_frames_volume_convolution(
                                **field_data,
                                **psf_data,
                                asframes=True,
                                as_mask=True,
                            )
                        beads = None
                        noise = False
                    elif convolution_type == "raw_volume":
                            images = conv.generate_frames_volume_convolution(
                                **field_data,
                                **psf_data,
                                asframes=False,
                            )
                            beads = None
                    elif convolution_type == "raw_volume_activation":
                            images = conv.generate_frames_volume_convolution(
                                **field_data,
                                **psf_data,
                                asframes=False,
                                activation_only=True
                            )
                            beads = None
                    else:
                        images, beads = self.images_by_convolutions(
                            convolution_type, **simparams
                        )
                    writing_notes_fluo = writing_notes_fluo + emission_notes
                    # print(f"max and min only photons: {np.max(images)}, {np.min(images)}")
                    # Beta implementation to save images before detection
                    if save:
                        self._save_timeseries_with_beads(images, beads, writing_notes_fluo)

                    # wrap up the images from a single fluorophore in the channel
                    output_per_fluoname[fluo] = dict(images=images, beads=beads)
                if convolution_type not in ["raw_volume", "raw_volume_activation"]:
                    timeseries_noiseless, beadstack_noiseless = self._add_fluorophore_signals(output_per_fluoname)
                    # # # Up to here only the photon information on arrival
                    timeseries_noise = None
                    beadstack_noise = None
                    if noise:
                        print("Adding noise")
                        timeseries_noise = np.zeros(
                            shape = timeseries_noiseless.shape,
                            dtype=np.float32)
                        timeseries_noise = self._crop_negative(timeseries_noiseless)
                        # print(np.max(timeseries), np.min(timeseries))
                        timeseries_noise = self.add_detector_noise(modality, timeseries_noise)
                        if beadstack_noiseless is not None:
                            beadstack_noise = np.zeros(
                                shape=beadstack_noiseless.shape,
                                dtype=np.float32)
                            beadstack_noise = self.add_detector_noise(modality, beadstack_noiseless)
                        else:
                            beadstack_noise = None
                    else:
                        timeseries_noise = copy.deepcopy(timeseries_noiseless)
                        beadstack_noise = copy.deepcopy(beadstack_noiseless)
                    if save:
                        # gt_positions = self.generate_ground_truth_positions(groundtruth_emitters)
                        timeseries_noise = self._crop_negative(timeseries_noise)
                        beadstack_noise = self._crop_negative(beadstack_noise)
                        writing_notes_fluo = writing_notes_fluo + "_withNoise_"

                        self._save_timeseries_with_beads(timeseries_noise, beadstack_noise, writing_notes_fluo)
                    output_per_channel[ch] = timeseries_noise
                    output_per_channel_noiseless[ch] = timeseries_noiseless
                    beads_per_channel[ch] = beadstack_noise
                    beads_per_channel_noiseless[ch] = beadstack_noiseless
                    #return timeseries_noise, beadstack_noise, timeseries_noiseless, beadstack_noiseless
                else:
                    ch = convolution_type
                    images_noiseless = None
                    beadstack_noiseless = None
                    output_per_channel[ch] = images
                    output_per_channel_noiseless[ch] = images_noiseless
                    beads_per_channel[ch] = beads
                    beads_per_channel_noiseless[ch] = beadstack_noiseless
                    #return images, beads, images_noiseless, beadstack_noiseless
        return (output_per_channel, beads_per_channel, output_per_channel_noiseless, beads_per_channel_noiseless)

    def generate_modality_mask(self, modality=None, channel = "ch0"):
        _1, _2, vsample_binay_positions, _3 = self.generate_imaging(
            modality=modality, 
            masks="mask")
        particle_positions = np.argwhere(vsample_binay_positions[channel][0] > 0)
        if particle_positions.shape[0] > 0:
            psf_lateral_resolution = self.modalities[modality]["psf"]["std_devs"][0]
            pixelsize_detection = self.modalities[modality]["detector"]["pixelsize"] * 1000
            particle_size = self.particle_size * 100
            particle_and_width = (psf_lateral_resolution + particle_size) * 10
            modality_psf_width_px = int(np.ceil(particle_and_width/pixelsize_detection))
            print(f"width: {modality_psf_width_px}, p_size: {particle_and_width}, psf_lateral_resolution: {psf_lateral_resolution}, {pixelsize_detection}")
            mask_with_psf = np.zeros(shape=vsample_binay_positions[channel][0].shape)
            max_x, max_y = vsample_binay_positions[channel][0].shape
            for i in range(particle_positions.shape[0]):
                x, y = particle_positions[i]
                low_x = np.floor(x - modality_psf_width_px)
                if low_x < 0:
                    low_x = 0
                high_x = np.ceil(x + modality_psf_width_px)
                if high_x > max_x:
                    high_x = max_x
                low_y = np.floor(y - modality_psf_width_px)
                if low_y < 0:
                    low_y = 0
                high_y = np.ceil(y +  modality_psf_width_px)
                if high_x > max_y:
                    high_x = max_y
                mask_with_psf[int(low_x):int(high_x), int(low_y):int(high_y)] = 1
            return mask_with_psf
        else:
            return np.zeros(shape=vsample_binay_positions[channel][0].shape)

    def write_ground_truth_positions(
        self, emitters, header=None, notes="", no_emitters=False
    ):
        # first, create the list of lines
        textlines = []
        # header = "x [nm],y [nm],z [nm]"
        textlines.append(header)
        if no_emitters:
            self.write_text(textlines, notes)
        else:
            for i in range(emitters.shape[0]):
                string_ = generate_coordinate_textline(
                    emitters[i, 0], emitters[i, 1], emitters[i, 2]
                )
                textlines.append(string_)
            self.write_text(textlines, notes)

    def _crop_negative(self, stack):
        if stack is None:
            return None
        else:
            if np.min(stack) < 0:
                # print("negative numbers found. Setting to zero")
                offset = np.ones(np.shape(stack)) * (-np.min(stack))
                stack = np.add(stack, offset)
            return stack

    def _add_fluorophore_signals(self, output_per_fluoname: dict):
        """
        Dictionary has the fluorophores as keys, each fluorophore
        will have the timeseries and can also have the stack of beads
        if exists
        """
        fluorophores = list(output_per_fluoname.keys())
        shp_ims = output_per_fluoname[fluorophores[0]]["images"].shape
        timeseries = np.zeros((shp_ims))
        for fl in fluorophores:
            both_images = np.array([timeseries, output_per_fluoname[fl]["images"]])
            timeseries = np.sum(both_images, axis=0)
        if output_per_fluoname[fluorophores[0]]["beads"] is not None:
            shp_bds = output_per_fluoname[fluorophores[0]]["beads"].shape
            beadstack = np.zeros((shp_bds))
            for fl in fluorophores:
                both_beads = np.array([beadstack, output_per_fluoname[fl]["beads"]])
                beadstack = np.sum(both_beads, axis=0)
        else:
            beadstack = None
        return timeseries, beadstack

    def add_detector_noise(self, modality, stack):
        """
        Add detector noise to an image stack according to the modality's noise model.

        Parameters
        ----------
        modality : str
            Name of the modality.
        stack : numpy.ndarray
            Image stack to which noise will be added.

        Returns
        -------
        numpy.ndarray
            Image stack with noise added.
        """

        # consider all the parameters of the detection system and corrupt
        # the photon data
        noise_params = self.modalities[modality]["detector"]["noise_model"]
        noise_order = self.modalities[modality]["detector"]["noise_order"]
        for noise_type in noise_order:
            # print(noise_type)
            stack = add_image_noise(noise_type, stack, **noise_params[noise_type])
            # includes the processes from collecting the photons, multiplying
            # and all possible source of noise
            # plus the ADU conversion factor
        # Efectively is the digitalisation part and sets the saturation level
        stack = self._adjust_to_pixel_depth(modality, stack)
        return stack

    def _adjust_to_pixel_depth(self, modality, stack):
        bits = self.modalities[modality]["detector"]["bits_pixel"]
        saturation = (2**bits) - 1
        stack[stack > saturation] == saturation
        return stack

    def _save_timeseries_with_beads(self, timeseries_stack, beads_stack, notes):
        notes_images = notes + "_timeseries_photons"
        self.write_tiff_stack(timeseries_stack, notes_images)
        if beads_stack is not None:
            notes_beads = notes + "_calibrationbeads_photons"
            self.write_tiff_stack(beads_stack, notes_beads)

    def images_by_convolutions(
        self,
        convolution_type,
        field_data,
        psf_data,
        nbeads,
        photons_per_bead,
        psf_projection_depth=1,
    ):
        if convolution_type == "2Dlookup":
            # generate convolutions plane by plane then collapse
            downsampled_images, upsampled_images = conv.generate_frames_projection_conv(
                **field_data, **psf_data
            )
            # print("GENERATING BEADS")
            if nbeads >= 1:
                downsampled_beads, upsampled_beads = (
                    conv.generate_beads_projection_conv(
                        **field_data,
                        **psf_data,
                        nbeads=nbeads,
                        bead_photons=photons_per_bead,
                    )
                )
            else:
                downsampled_beads = None
        elif convolution_type == "2Dlookup_optim":
            emitters, images, downsampled_images = (
                conv.generate_frames_projection_conv_optimised2(
                    **field_data, **psf_data
                )
            )
            if nbeads >= 1:
                downsampled_beads, upsampled_beads = (
                    conv.generate_beads_projection_conv(
                        **field_data,
                        **psf_data,
                        nbeads=nbeads,
                        bead_photons=photons_per_bead,
                    )
                )
            else:
                downsampled_beads = None
        elif convolution_type == "3Dvolume":
            # generate convolutions using the whole volume
            downsampled_images = conv.generate_frames_volume_convolution(
                **field_data,
                **psf_data,
                asframes=True,
                psf_projection_depth=psf_projection_depth,
            )
            if nbeads >= 1:
                downsampled_beads = conv.generate_beads_frames_volume_convolution(
                    **field_data,
                    **psf_data,
                    nbeads=nbeads,
                    bead_photons=photons_per_bead,
                    asframes=True,
                )
            else:
                downsampled_beads = None
        return (
            downsampled_images,
            downsampled_beads,
        )  # change at will if needed to get the original images

    def get_emitters_in_ROI(self, fluoname: str, masks=False):
        """
        Get emitters of a given fluorophore that are within the ROI.

        Parameters
        ----------
        fluoname : str
            Name of the fluorophore.

        Returns
        -------
        numpy.ndarray
            Array of emitter coordinates within the ROI.
        """
        # get limits of ROI in xyz
        # to match the previous implementation
        # the limits must be from 0 to maxX in microns
        ranges = self.get_roi_params("ranges")
        if masks:
            points = self.particle_positions
        else:
            points = self._get_emitters_by_fluorophorename(fluoname)
        # print(ranges, points)
        rangesT = np.array(ranges).T
        roi_corners = [rangesT[0].tolist(), rangesT[1].tolist()]
        logical = [inCube(X, roi_corners) for X in points]
        emitters_in_ROI = points[logical, :]
        # up to here the points are only the ones contined
        # in the ROI, but we still need to offset the coordinates
        return copy.copy(emitters_in_ROI)

    def calculate_photons_per_frame(
        self, modality, fluo, n_emitters, nframes, exp_time=1, **kwargs
    ):
        """
        Calculate the number of photons per frame for emitters.

        Parameters
        ----------
        modality : str
            Name of the modality.
        fluo : str
            Name of the fluorophore.
        n_emitters : int
            Number of emitters.
        nframes : int
            Number of frames.
        exp_time : float, optional
            Exposure time per frame. Default is 1.
        **kwargs
            Additional keyword arguments.

        Returns
        -------
        photon_frames : numpy.ndarray
            Array of photon counts per frame.
        emission_notes : str
            Notes about the emission model used.
        """
        # which model to use
        if n_emitters == 0:
            photon_frames = np.repeat(0, nframes)
            emission_notes = "noemitters"
        else:
            emission = self.modalities[modality]["emission"]
            if emission == "blinking":
                kinetics = dict(self.fluorophore_params[fluo]["blinking"])
                # print(nframes, n_emitters,kinetics)
                photon_frames = self._generate_photons_blinking_modality(
                    nframes, n_emitters, **kinetics
                )
                emission_notes = dictionary2string(kinetics)
            else: #if emission == "constant":
                photons_per_second = self.fluorophore_params[fluo]["photons_per_second"]
                photons_per_frame = exp_time * photons_per_second
                print(f"Average number of photons per frame: {photons_per_frame}")
                photon_frames = self._generate_constant_emission_modality(
                    nframes, n_emitters, photons_per_frame
                )
                emission_notes = "constant_emission_"
        return photon_frames, emission_notes

    def _generate_photons_blinking_modality(self, nframes, nemitters, **kwargs):
        """
        kwargs is intended to hold the kinetic parameters and
        photon budget from the emitter species
        """
        # simulate blinking traces for the N emitters
        # kinetic parameters are in seconds
        tcs = blinking_2states_1bleach_bulk(nframes, nemitters, **kwargs)
        blinking_traces = binary_trace(tcs)  ##
        photons_array = np.random.poisson(
            kwargs["photons_per_blink"], nemitters * nframes
        ).reshape(nemitters, nframes)
        photons_frames = np.multiply(photons_array, blinking_traces)
        return photons_frames

    def _generate_constant_emission_modality(
        self, nframes, nemitters, photons_per_second, exposure_time=1, **kwargs
    ):
        exposure_photons = photons_per_second * exposure_time
        photons_frames = np.random.poisson(
            exposure_photons, nemitters * nframes
        ).reshape(nemitters, nframes)
        return photons_frames

    def _homogenise_scales4convolution_modality(
        self, modality, emitters_in_channel, photons_frames, **kwargs
    ):
        """
        Compiles all data necesary to carry on the generation of frames
        considering the input coordinates
        the calclated photons per frame
        and the PSF of the input modality
        and expresses all dimensions in the dimensions of the PSF

        For example, if the image generator scale is in micrometers
        all the coordinates will be set to nanometers, which is the expected
        scale for the pixelsize of the PSF

        Ideally, the resulting dictionary is passed to a custom convolution
        function

        """
        # check if scaling is needed
        psf_scale = self.modalities[modality]["psf"]["scale"]
        # print(f"psf scale: psf_scale")
        scaling = math.ceil(self.get_roi_params("scale") / psf_scale)
        # print(f"scaling for converting field data for convolution: {scaling}")
        # define field values
        scaling_used = scaling
        field_size = self.modalities[modality]["detector"][
            "image_size"
        ]  # in pixels xyz
        field_pixelsizeXY = (
            self.modalities[modality]["detector"]["pixelsize"] * scaling
        )  # detector pixelsize
        field_coordinates = emitters_in_channel * scaling
        field_zfocus = self.get_roi_params("focus_plane") * scaling
        # ALL PARAMETERS REGARDING FIELD ARE NOW IN PSF SCALE UNITS
        field_scale = psf_scale
        field_data = dict(
            field_scale=field_scale,
            field_size=field_size,
            field_pixelsizeXY=field_pixelsizeXY,
            field_coordinates=field_coordinates,
            field_zfocus=field_zfocus,
            photons_frames=photons_frames,
        )
        # psf parameters
        psf_scale = psf_scale
        psf_array = self.modalities[modality]["psf"]["psf_stack"]
        psf_zstep = self.modalities[modality]["psf"]["voxelsize"][
            2
        ]  # dimensions are assumed to be xyz
        psf_pixelsizeXY = self.modalities[modality]["psf"]["voxelsize"][
            0
        ]  # equal pixelsize laterally
        psf_focus_slice = int(self.modalities[modality]["psf"]["focus_plane"])
        psf_data = dict(
            psf_scale=psf_scale,
            psf_array=psf_array,
            psf_zstep=psf_zstep,
            psf_pixelsizeXY=psf_pixelsizeXY,
            psf_focus_slice=psf_focus_slice,
        )
        # print(f"Matching PSF and coordinates scales")
        return field_data, psf_data

    def _get_emitters_by_fluorophorename(self, fluo):
        return np.array(self.emitters_by_fluorophore[fluo])

    def set_writing_directory(self, saving_dir):
        """
        Set the directory for saving output files.

        Parameters
        ----------
        saving_dir : str
            Path to the directory for saving files.
        """
        self.writing_dir = saving_dir

    def write_tiff_stack(self, image_stk, notes):
        """
        Write an image stack to a TIFF file.

        Parameters
        ----------
        image_stk : numpy.ndarray
            Image stack to save.
        notes : str
            Notes or filename suffix for the output file.
        """
        path = self.writing_dir
        write_tif(image_stk, path, notes)

    def write_text(self, list_of_lines, notes):
        """
        Write a list of text lines to a file.

        Parameters
        ----------
        list_of_lines : list of str
            Lines to write.
        notes : str
            Notes or filename suffix for the output file.
        """
        # text is written line by line
        path = self.writing_dir
        write_txt(list_of_lines, path, notes)

    # methods for visualisation
    def show_field(
        self,
        fluo_type="all",
        view_init=[30, 0, 0],
        initial_pos=True,
        reference_pt=False,
        axesoff=False,
        emitters_plotsize=None,
        return_fig = False
    ):
        """
        Visualize the emitter field and ROI in 3D.

        Parameters
        ----------
        fluo_type : str, optional
            Fluorophore type to display. Default is "all".
        view_init : list of int, optional
            Initial view angles [elev, azim, roll]. Default is [30, 0, 0].
        initial_pos : bool, optional
            If True, show initial positions. Default is True.
        reference_pt : bool, optional
            If True, show the reference point. Default is False.
        axesoff : bool, optional
            If True, hide axes. Default is False.
        return_fig : bool, optional
            If True, return the matplotlib figure. Default is False.

        Returns
        -------
        matplotlib.figure.Figure or None
            The figure if return_fig is True, otherwise None.
        """
        # FOCUS
        roi_ranges = self.get_roi_params("ranges")
        z_focus = self.get_roi_params("focus_plane")
        visualisation_scale = 1e-9  # needed for meshgrid
        factor = self.get_roi_params("scale") / visualisation_scale
        xrang = np.array(roi_ranges[0]) * factor
        yrang = np.array(roi_ranges[1]) * factor
        plot_scale = self.get_roi_params("scale") / factor
        # print(f"Showing field on scale: {plot_scale} meters")
        # print(f"dimension ranges in XY pane: {xrang, yrang}")
        # print(f"position of focus {z_focus}")
        xx, yy = np.meshgrid(
            range(int(xrang[0]), int(xrang[1])), range(int(yrang[0]), int(yrang[1]))
        )
        zz = yy * 0 + (z_focus * factor)
        fig = plt.figure()
        ax = fig.add_subplot(111, projection="3d")
        ax.plot_surface(xx, yy, zz, alpha=0.2)
        # show ROI reference point
        if reference_pt:
            ref_pt = self.get_absolute_reference_point() * factor
            add_ax_scatter(ax, format_coordinates(ref_pt))
        # EMITTERS PER FLUOROPHORE SPECIES
        if self.emitters_by_fluorophore is not None:
            for fname, coords in self.emitters_by_fluorophore.items():
                # print(fname)
                if emitters_plotsize is not None:
                    self.plotting_params[fname]["plotsize"] = emitters_plotsize
                add_ax_scatter(
                    ax,
                    format_coordinates(coords * factor, **self.plotting_params[fname]),
                )
        ax.set_box_aspect(
            [ub - lb for lb, ub in (getattr(ax, f"get_{a}lim")() for a in "xyz")]
        )
        ax.view_init(elev=view_init[0], azim=view_init[1], roll=view_init[2])
        if axesoff:
            ax.set_axis_off()
        if return_fig:
            return fig


    def get_modality_psf_stack(self, modality: str):
        """
        Get the PSF stack for a given modality.

        Parameters
        ----------
        modality : str
            Name of the modality.

        Returns
        -------
        numpy.ndarray
            PSF stack for the modality.

        Raises
        ------
        ValueError
            If the modality is not found.
        """
        if modality in self.modalities.keys():
            return self.modalities[modality]["psf"]["psf_stack"]
        else:
            raise ValueError(f"Modality {modality} not found in modalities")


from vlab4mic.utils.data_format.structural_format import label_builder_format
import pytest
import numpy as np
from vlab4mic import workflows

structural_integrity_values = np.linspace(0, 1, 3)


@pytest.mark.parametrize("structural_integrity", structural_integrity_values)
def test_add_structural_integrity(structural_integrity, experiment_7r5k_base):
    assert experiment_7r5k_base.generators_status("particle")
    experiment_7r5k_base.particle.add_structural_integrity(
        eps1=300,
        xmer_neigh_distance=600,
        integrity=structural_integrity,
    )
    if structural_integrity == 1:
        assert experiment_7r5k_base.particle.structural_integrity_target_normals is None
    else:
        assert experiment_7r5k_base.particle.structural_integrity_target_normals is not None
